from __future__ import print_function, absolute_import, division
from argparse import ArgumentDefaultsHelpFormatter

__author__ = 'muneeb'


def func(args, parser):
    # delay import of the rest of the module to improve `osprey -h` performance
    from ..execute_currentbest import execute
    execute(args, parser)


def configure_parser(sub_parsers):
    help = 'Get parameters for the current best model'
    p = sub_parsers.add_parser('current_best', description=help, help=help,
                               formatter_class=ArgumentDefaultsHelpFormatter)
    p.add_argument('config', help='Path to worker config file (yaml)')

    p.set_defaults(func=func)



from __future__ import print_function, absolute_import, division

from os.path import join, exists
from pkg_resources import resource_filename


def execute(args, parser):
    if args.template == 'msmbuilder':
        fn = resource_filename('osprey', join('data',
                               'msmbuilder_skeleton_config.yaml'))
    elif args.template == 'sklearn':
        fn = resource_filename('osprey', join('data',
                               'sklearn_skeleton_config.yaml'))
    else:
        raise RuntimeError('unknown template: %s' % args.template)

    if exists(args.filename):
        raise RuntimeError('file already exists: %s' % args.filename)

    print("\033[92mcreate\033[0m  {0:s}".format(args.filename))

    with open(args.filename, 'wb') as out:
        with open(fn, 'rb') as inp:
            out.write(inp.read())



from __future__ import print_function, absolute_import, division

import os
import sys
import time
import signal
import traceback
from socket import gethostname
from getpass import getuser
from datetime import datetime

from six import iteritems
from six.moves import cStringIO
from sqlalchemy import func
from sklearn.base import clone, BaseEstimator
import numpy as np

from . import __version__
from .config import Config
from .trials import Trial
from .fit_estimator import fit_and_score_estimator
from .utils import Unbuffered, format_timedelta, current_pretty_time
from .utils import is_msmbuilder_estimator, num_samples


def execute(args, parser):
    start_time = datetime.now()
    sys.stdout = Unbuffered(sys.stdout)
    # Load the config file and extract the fields
    print_header()

    config = Config(args.config)
    estimator = config.estimator()
    searchspace = config.search_space()
    strategy = config.strategy()
    config_sha1 = config.sha1()
    scoring = config.scoring()
    project_name = config.project_name()

    if is_msmbuilder_estimator(estimator):
        print_msmbuilder_version()

    print('\nLoading dataset...\n')
    X, y = config.dataset()
    print('Dataset contains %d element(s) with %s labels'
          % (num_samples(X), 'out' if y is None else ''))
    print('The elements have shape: [%s' %
          ', '.join([str(X[i].shape)
                     if isinstance(X[i], (np.ndarray, np.generic))
                     else '(%s,)' % num_samples(X[i])
                     for i in range(min(num_samples(X), 20))]), end='')
    print(', ...]' if (num_samples(X) > 20) else ']')
    print('Instantiated estimator:')
    print('  %r' % estimator)
    print(searchspace)

    # set up cross-validation
    cv = config.cv(X, y)

    statuses = [None for _ in range(args.n_iters)]

    # install a signal handler to print the footer before exiting
    # from sigterm (e.g. PBS job kill)
    def signal_hander(signum, frame):
        print_footer(statuses, start_time, signum)
        sys.exit(1)
    signal.signal(signal.SIGTERM, signal_hander)

    for i in range(args.n_iters):
        print('\n' + '-'*70)
        print('Beginning iteration %50s' % ('%d / %d' % (i+1, args.n_iters)))
        print('-'*70)

        trial_id, params = initialize_trial(
            strategy, searchspace, estimator, config_sha1=config_sha1,
            project_name=project_name,
            sessionbuilder=config.trialscontext)

        s = run_single_trial(
            estimator=estimator, params=params, trial_id=trial_id,
            scoring=scoring, X=X, y=y, cv=cv,
            sessionbuilder=config.trialscontext)

        statuses[i] = s

    print_footer(statuses, start_time)


def initialize_trial(strategy, searchspace, estimator, config_sha1,
                     project_name, sessionbuilder):

    with sessionbuilder() as session:
        # requery the history ever iteration, because another worker
        # process may have written to it in the mean time
        history = [[t.parameters, t.test_scores, t.status]
                   for t in session.query(Trial).all()
                   if t.project_name == project_name]

        print('History contains: %d trials' % len(history))
        print('Choosing next hyperparameters with %s...' % strategy.short_name)
        start = time.time()
        params = strategy.suggest(history, searchspace)
        print('  %r' % params)
        print('(%s took %.3f s)\n' % (strategy.short_name,
                                      time.time() - start))
        assert len(params) == searchspace.n_dims

        # make sure we get _all_ the parameters, including defaults on the
        # estimator class, to save in the database
        params = clone(estimator).set_params(**params).get_params()
        params = dict((k, v) for k, v in iteritems(params)
                      if not isinstance(v, BaseEstimator) and
                      (k != 'steps'))

        t = Trial(status='PENDING', parameters=params, host=gethostname(),
                  user=getuser(), started=datetime.now(),
                  config_sha1=config_sha1)
        session.add(t)
        session.commit()
        trial_id = t.id

    return trial_id, params


def run_single_trial(estimator, params, trial_id, scoring, X, y, cv,
                     sessionbuilder):

    status = None

    try:
        score = fit_and_score_estimator(
            estimator, params, cv=cv, scoring=scoring, X=X, y=y, verbose=1)
        with sessionbuilder() as session:
            trial = session.query(Trial).get(trial_id)
            trial.mean_test_score = score['mean_test_score']
            trial.mean_train_score = score['mean_train_score']
            trial.test_scores = score['test_scores']
            trial.train_scores = score['train_scores']
            trial.n_test_samples = score['n_test_samples']
            trial.n_train_samples = score['n_train_samples']

            trial.status = 'SUCCEEDED'
            best_so_far = session.query(
                func.max(Trial.mean_test_score)).first()
            print('~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~')
            print('Success! Model score = %f' % trial.mean_test_score)
            print('(best score so far   = %f)' %
                  max(trial.mean_test_score, best_so_far[0]))
            print('~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~')
            trial.completed = datetime.now()
            trial.elapsed = trial.completed - trial.started
            session.commit()
            status = trial.status

    except Exception:
        buf = cStringIO()
        traceback.print_exc(file=buf)

        with sessionbuilder() as session:
            trial = session.query(Trial).get(trial_id)
            trial.traceback = buf.getvalue()
            trial.status = 'FAILED'
            print('-'*78, file=sys.stderr)
            print('Exception encountered while fitting model')
            print('-'*78, file=sys.stderr)
            traceback.print_exc(file=sys.stderr)
            print('-'*78, file=sys.stderr)
            session.commit()
            status = trial.status

    except (KeyboardInterrupt, SystemExit):
        with sessionbuilder() as session:
            trial = session.query(Trial).get(trial_id)
            trial.status = 'FAILED'
            session.commit()
            sys.exit(1)

    return status


def print_header():
    print('='*70)
    print('= osprey is a tool for machine learning '
          'hyperparameter optimization. =')
    print('='*70)
    print()
    print('osprey version:      %s' % __version__)
    print('time:                %s' % current_pretty_time())
    print('hostname:            %s' % gethostname())
    print('cwd:                 %s' % os.path.abspath(os.curdir))
    print('pid:                 %s' % os.getpid())
    print()


def print_msmbuilder_version():
    from msmbuilder.version import full_version as msmb_version
    from mdtraj.version import full_version as mdtraj_version
    print()
    print('msmbuilder version:  %s' % msmb_version)
    print('mdtraj version:      %s' % mdtraj_version)
    print()


def print_footer(statuses, start_time, signum=None):
    n_successes = sum(s == 'SUCCEEDED' for s in statuses)
    elapsed = format_timedelta(datetime.now() - start_time)
    print()

    if signum is not None:
        sigmap = dict((k, v) for v, k in iteritems(signal.__dict__)
                      if v.startswith('SIG'))
        signame = sigmap.get(signum, 'Unknown')
        print('== osprey worker received signal %s!' % signame,
              file=sys.stderr)
        print('== exiting immediately.', file=sys.stderr)

    print('%d/%d models fit successfully.' % (n_successes, len(statuses)))
    print('time:         %s' % current_pretty_time())
    print('elapsed:      %s.' % elapsed)
    print('osprey worker exiting.')



from __future__ import print_function, absolute_import, division

from .trials import Trial
from .config import Config
from .plot import plot_1, plot_2, plot_3, plot_4

try:
    import bokeh.plotting as bk
    from bokeh.io import vplot
except ImportError:
    raise RuntimeError(
        'This command requires the Bokeh library (http://bokeh.pydata.org/) '
        'version >=0.10.0.\n\n    $ conda install bokeh  # (recommended)\n'
        'or\n    $ pip install bokeh')

PLOTS = [plot_4, plot_1, plot_2, plot_3]


def execute(args, parser):
    config = Config(args.config, verbose=False)
    with config.trialscontext() as session:
        q = (session.query(Trial)
             .filter(Trial.status == 'SUCCEEDED')
             .order_by(Trial.started))
        data = [curr.to_dict() for curr in q.all()]

    bk.output_file(args.filename, title='osprey')

    plots = []
    ss = config.search_space()
    for plot in PLOTS:
        plt = plot(data, ss)
        if plt is not None:
            plt = plt if isinstance(plt, list) else [plt]
            plots.extend(plt)

    p = vplot(*plots)
    if args.browser:
        bk.show(p)
    else:
        bk.save(p)



from __future__ import print_function, absolute_import, division
import sys
import inspect
import socket

import numpy as np
from sklearn.utils import check_random_state
from sklearn.grid_search import ParameterGrid
try:
    from hyperopt import (Trials, tpe, fmin, STATUS_OK, STATUS_RUNNING,
                          STATUS_FAIL)
except ImportError:
    # hyperopt is optional, but required for hyperopt_tpe()
    pass

try:
    from GPy import kern
    from GPy.kern import RBF, Fixed, Bias
    from GPy.util.linalg import tdot
    from GPy.models import GPRegression
    from scipy.optimize import minimize
except:
    # GPy is optional, but required for gp
    GPRegression = kern = minimize = None
    pass
from .search_space import EnumVariable

DEFAULT_TIMEOUT = socket._GLOBAL_DEFAULT_TIMEOUT


class BaseStrategy(object):
    short_name = None

    def suggest(self, history, searchspace):
        """
        Parameters
        ----------
        history : list of 3-tuples
            History of past function evaluations. Each element in history
            should be a tuple `(params, score, status)`, where `params` is a
            dict mapping parameter names to values
        searchspace : SearchSpace
            Instance of search_space.SearchSpace
        random_state :i nteger or numpy.RandomState, optional
            The random seed for sampling. If an integer is given, it fixes the
            seed. Defaults to the global numpy random number generator.

        Returns
        -------
        new_params : dict
        """
        raise NotImplementedError()


class RandomSearch(BaseStrategy):
    short_name = 'random'

    def __init__(self, seed=None):
        self.seed = seed

    def suggest(self, history, searchspace):
        """Randomly suggest params from searchspace.
        """
        return searchspace.rvs(self.seed)


class HyperoptTPE(BaseStrategy):
    short_name = 'hyperopt_tpe'

    def __init__(self, seed=None):
        self.seed = seed

    def suggest(self, history, searchspace):
        """
        Suggest params to maximize an objective function based on the
        function evaluation history using a tree of Parzen estimators (TPE),
        as implemented in the hyperopt package.

        Use of this function requires that hyperopt be installed.
        """
        # This function is very odd, because as far as I can tell there's
        # no real documented API for any of the internals of hyperopt. Its
        # execution model is that hyperopt calls your objective function
        # (instead of merely providing you with suggested points, and then
        # you calling the function yourself), and its very tricky (for me)
        # to use the internal hyperopt data structures to get these predictions
        # out directly.

        # so they path we take in this function is to construct a synthetic
        # hyperopt.Trials database which from the `history`, and then call
        # hyoperopt.fmin with a dummy objective function that logs the value
        # used, and then return that value to our client.

        # The form of the hyperopt.Trials database isn't really documented in
        # the code -- most of this comes from reverse engineering it, by
        # running fmin() on a simple function and then inspecting the form of
        # the resulting trials object.
        if 'hyperopt' not in sys.modules:
            raise ImportError('No module named hyperopt')

        random = check_random_state(self.seed)
        hp_searchspace = searchspace.to_hyperopt()

        trials = Trials()
        for i, (params, scores, status) in enumerate(history):
            if status == 'SUCCEEDED':
                # we're doing maximization, hyperopt.fmin() does minimization,
                # so we need to swap the sign
                result = {'loss': -np.mean(scores), 'status': STATUS_OK}
            elif status == 'PENDING':
                result = {'status': STATUS_RUNNING}
            elif status == 'FAILED':
                result = {'status': STATUS_FAIL}
            else:
                raise RuntimeError('unrecognized status: %s' % status)

            # the vals key in the trials dict is basically just the params
            # dict, but enum variables (hyperopt hp.choice() nodes) are
            # different, because the index of the parameter is specified
            # in vals, not the parameter itself.

            vals = {}
            for var in searchspace:
                if isinstance(var, EnumVariable):
                    # get the index in the choices of the parameter, and use
                    # that.
                    matches = [i for i, c in enumerate(var.choices)
                               if c == params[var.name]]
                    assert len(matches) == 1
                    vals[var.name] = matches
                else:
                    # the other big difference is that all of the param values
                    # are wrapped in length-1 lists.
                    vals[var.name] = [params[var.name]]

            trials.insert_trial_doc({
                'misc': {
                    'cmd': ('domain_attachment', 'FMinIter_Domain'),
                    'idxs': dict((k, [i]) for k in hp_searchspace.keys()),
                    'tid': i,
                    'vals': vals,
                    'workdir': None},
                'result': result,
                'tid': i,
                # bunch of fixed fields that hyperopt seems to require
                'owner': None, 'spec': None, 'state': 2, 'book_time': None,
                'exp_key': None, 'refresh_time': None, 'version': 0
                })

        trials.refresh()
        chosen_params_container = []

        def mock_fn(x):
            # http://stackoverflow.com/a/3190783/1079728
            # to get around no nonlocal keywork in python2
            chosen_params_container.append(x)
            return 0

        fmin(fn=mock_fn, algo=tpe.suggest, space=hp_searchspace, trials=trials,
             max_evals=len(trials.trials)+1,
             **self._hyperopt_fmin_random_kwarg(random))
        chosen_params = chosen_params_container[0]

        return chosen_params

    @staticmethod
    def _hyperopt_fmin_random_kwarg(random):
        if 'rstate' in inspect.getargspec(fmin).args:
            # 0.0.3-dev version uses this argument
            kwargs = {'rstate': random, 'allow_trials_fmin': False}
        elif 'rseed' in inspect.getargspec(fmin).args:
            # 0.0.2 version uses different argument
            kwargs = {'rseed': random.randint(2**32-1)}
        return kwargs


class GP(BaseStrategy):
    short_name = 'gp'

    def __init__(self, seeds=1, max_feval=5E4, max_iter=1E5):
        self.seeds = seeds
        self.max_feval = max_feval
        self.max_iter = max_iter
        self.model = None
        self.n_dims = None
        self.kernel = None
        self._kerns = None
        self._kernf = None
        self._kernb = None

    def _create_kernel(self, V):
        self._kerns = [RBF(1, ARD=True, active_dims=[i])
                       for i in range(self.n_dims)]
        self._kernf = Fixed(self.n_dims, tdot(V))
        self._kernb = Bias(self.n_dims)
        self.kernel = np.sum(self._kerns) + self._kernf + self._kernb

    def _fit_model(self, X, Y):
        model = GPRegression(X, Y, self.kernel)
        model.optimize(messages=False, max_f_eval=self.max_feval)
        self.model = model

    def _get_random_point(self):
        return np.array([np.random.uniform(low=0., high=1.)
                         for i in range(self.n_dims)])

    def _optimize(self, init=None):
        if not init:
            init = self._get_random_point()

        def z(x):
            y = x.copy().reshape(-1, self.n_dims)
            s, v = self.model.predict(y, kern=(np.sum(self._kerns).copy() +
                                               self._kernb.copy()))
            return -(s+v).flatten()

        return minimize(z, init, bounds=self.n_dims*[(0., 1.)],
                        options={'maxiter': self.max_iter, 'disp': 0}).x

    def _get_data(self, history, searchspace):
        X = []
        Y = []
        V = []
        ignore = []
        for param_dict, scores, status in history:
            # transform points into the GP domain. This invloves bringing
            # int and enum variables to floating point, etc.
            if status == 'FAILED':
                # not sure how to deal with these yet
                continue

            point = searchspace.point_to_gp(param_dict)
            if status == 'SUCCEEDED':
                X.append(point)
                Y.append(np.mean(scores))
                V.append(np.var(scores))

            elif status == 'PENDING':
                ignore.append(point)
            else:
                raise RuntimeError('unrecognized status: %s' % status)

        return (np.array(X).reshape(-1, self.n_dims),
                np.array(Y).reshape(-1, 1),
                np.array(V).reshape(-1, 1),
                np.array(ignore).reshape(-1, self.n_dims))

    def _from_gp(self, result, searchspace):

        # Note that GP only deals with float-valued variables, so we have
        # a transform step on either side, where int and enum valued variables
        # are transformed before calling gp, and then the result suggested by
        # GP needs to be reverse-transformed.
        out = {}
        for gpvalue, var in zip(result, searchspace):
            out[var.name] = var.point_from_gp(float(gpvalue))

        return out

    def _is_within(self, point, X, tol=1E-2):
        if True in (np.sqrt(((point - X)**2).sum(axis=0)) <= tol):
            return True
        return False

    def suggest(self, history, searchspace, max_tries=5):
        if not GPRegression:
            raise ImportError('No module named GPy')
        if not minimize:
            raise ImportError('No module named SciPy')

        if len(history) < self.seeds:
            return RandomSearch().suggest(history, searchspace)

        self.n_dims = searchspace.n_dims

        X, Y, V, ignore = self._get_data(history, searchspace)
        self._create_kernel(V)
        self._fit_model(X, Y)

        suggestion = self._optimize()

        if suggestion in ignore or self._is_within(suggestion, X):
            return RandomSearch().suggest(history, searchspace)

        return self._from_gp(suggestion, searchspace)


class GridSearch(BaseStrategy):
    short_name = 'grid'

    def __init__(self):
        self.param_grid = None
        self.current = -1

    def suggest(self, history, searchspace):
        # Convert searchspace to param_grid
        if self.param_grid is None:
            if not all(isinstance(v, EnumVariable) for v in searchspace):
                raise RuntimeError("GridSearchStrategy is defined only for all-enum search space")

            self.param_grid = ParameterGrid(dict((v.name, v.choices) for v in searchspace))

        # NOTE: there is no way of signaling end of parameters to be searched against
        # so user should pick correctly number of evaluations
        self.current += 1
        return self.param_grid[self.current % len(self.param_grid)]




from __future__ import print_function, absolute_import, division
import json
import time
import random
from datetime import datetime, timedelta

from six import iteritems

from sqlalchemy.pool import NullPool
from sqlalchemy.exc import OperationalError
from sqlalchemy import Column, create_engine
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.types import (TypeDecorator, Text, Float, Integer, Enum,
                              DateTime, String, Interval)
from sqlalchemy.orm import Session
Base = declarative_base()

__all__ = ['Trial']


class JSONEncoded(TypeDecorator):
    impl = Text

    def process_bind_param(self, value, dialect):
        if value is not None:
            value = json.dumps(value)
        return value

    def process_result_value(self, value, dialect):
        if value is not None:
            value = json.loads(value)
        return value


class Trial(Base):
    __tablename__ = 'trials_v3'
    default_project_name = None

    id = Column(Integer, primary_key=True)
    project_name = Column(Text())
    status = Column(Enum('PENDING', 'SUCCEEDED', 'FAILED'))
    parameters = Column(JSONEncoded())

    mean_test_score = Column(Float)
    mean_train_score = Column(Float)
    train_scores = Column(JSONEncoded())
    test_scores = Column(JSONEncoded())
    n_train_samples = Column(JSONEncoded())
    n_test_samples = Column(JSONEncoded())

    started = Column(DateTime())
    completed = Column(DateTime())
    elapsed = Column(Interval())
    host = Column(String(512))
    user = Column(String(512))
    traceback = Column(Text())
    config_sha1 = Column(String(40))

    @classmethod
    def set_default_project_name(cls, name):
        cls.default_project_name = name

    def __init__(self, **kwargs):
        if 'project_name' not in kwargs:
            kwargs['project_name'] = self.default_project_name
        Base.__init__(self, **kwargs)

    def to_dict(self):
        item = {}
        for k, v in iteritems(self.__dict__):
            if k.startswith('_'):
                continue
            if isinstance(v, (datetime, timedelta)):
                v = str(v)
            item[k] = v
        return item


def make_session(uri, project_name, echo=False):
    Trial.set_default_project_name(project_name)
    engine = create_engine(uri, echo=echo, poolclass=NullPool)
    _create_all(Base, engine)
    session = Session(engine)
    return session


def _create_all(base, engine):
    # when multiple workers start up at the same time, they
    # can have a race condition in creating the DB.
    error = None
    for i in range(3):
        try:
            return base.metadata.create_all(engine)
        except OperationalError as e:
            time.sleep(random.random())
            error = e
    raise error



from __future__ import print_function, absolute_import, division

import csv
import json
from six.moves import cStringIO
from .config import Config
from .trials import Trial


def execute(args, parser):
    config = Config(args.config, verbose=False)

    session = config.trials()
    columns = Trial.__mapper__.columns

    if args.output == 'json':
        items = [curr.to_dict() for curr in session.query(Trial).all()]
        value = json.dumps(items)

    elif args.output == 'csv':
        buf = cStringIO()
        outcsv = csv.writer(buf)
        outcsv.writerow([column.name for column in columns])
        for curr in session.query(Trial).all():
            row = [getattr(curr, column.name) for column in columns]
            outcsv.writerow(row)
        value = buf.getvalue()

    print(value)
    return value



import warnings
from collections import OrderedDict

import numpy as np
from matplotlib import cm
from matplotlib.colors import rgb2hex
from sklearn.manifold import TSNE

try:
    import pandas as pd
    import bokeh.plotting as bk
    from bokeh.models import HoverTool
    from bokeh.models.sources import ColumnDataSource
except ImportError:
    raise RuntimeError(
        'This command requires the Bokeh library (http://bokeh.pydata.org/) '
        'version >=0.10.0.\n\n    $ conda install bokeh  # (recommended)\n'
        'or\n    $ pip install bokeh')

TOOLS = "pan,wheel_zoom,box_zoom,reset,hover"


def nonconstant_parameters(data):
    assert len(data) > 0
    df = pd.DataFrame([d['parameters'] for d in data])
    # http://stackoverflow.com/a/20210048/1079728
    filtered = df.loc[:, (df != df.ix[0]).any()]
    return filtered


def build_scatter_tooltip(x, y, tt, add_line=True, radius=3, title='My Plot',
                          xlabel='Iteration number', ylabel='Score'):
    p = bk.figure(title=title, tools=TOOLS)

    p.circle(
        x, y, radius=radius, source=ColumnDataSource(tt),
        fill_alpha=0.6, line_color=None)

    if add_line:
        p.line(x, y, line_width=2)

    xax, yax = p.axis
    xax.axis_label = xlabel
    yax.axis_label = ylabel

    cp = p
    hover = cp.select(dict(type=HoverTool))
    format_tt = [(s, '@%s' % s) for s in tt.columns]
    hover.tooltips = OrderedDict([("index", "$index")] + format_tt)
    return p


def plot_1(data, *args):
    """Plot 1. All iterations (scatter plot)"""
    df_all = pd.DataFrame(data)
    df_params = nonconstant_parameters(data)
    return build_scatter_tooltip(
        x=df_all['id'], y=df_all['mean_test_score'], tt=df_params,
        title='All Iterations')


def plot_2(data, *args):
    """Plot 2. Running best score (scatter plot)"""
    df_all = pd.DataFrame(data)
    df_params = nonconstant_parameters(data)
    x = [df_all['id'][0]]
    y = [df_all['mean_test_score'][0]]
    params = [df_params.loc[0]]
    for i in range(len(df_all)):
        if df_all['mean_test_score'][i] > y[-1]:
            x.append(df_all['id'][i])
            y.append(df_all['mean_test_score'][i])
            params.append(df_params.loc[i])
    return build_scatter_tooltip(
        x=x, y=y, tt=pd.DataFrame(params), title='Running best')


def plot_3(data, ss, *args):
    """t-SNE embedding of the parameters, colored by score
    """

    if len(data) <= 1:
        warnings.warn("Only one datapoint. Could not compute t-SNE embedding.")
        return None

    scores = np.array([d['mean_test_score'] for d in data])
    # maps each parameters to a vector of floats
    warped = np.array([ss.point_to_gp(d['parameters']) for d in data])

    # Embed into 2 dimensions with t-SNE
    X = TSNE(n_components=2).fit_transform(warped)

    e_scores = np.exp(scores)
    mine, maxe = np.min(e_scores), np.max(e_scores)
    color = (e_scores - mine) / (maxe - mine)
    mapped_colors = list(map(rgb2hex, cm.get_cmap('RdBu_r')(color)))

    p = bk.figure(title='t-SNE (unsupervised)', tools=TOOLS)

    df_params = nonconstant_parameters(data)
    df_params['score'] = scores
    p.circle(
        X[:, 0], X[:, 1], color=mapped_colors, radius=1,
        source=ColumnDataSource(df_params), fill_alpha=0.6,
        line_color=None)
    cp = p
    hover = cp.select(dict(type=HoverTool))
    format_tt = [(s, '@%s' % s) for s in df_params.columns]
    hover.tooltips = OrderedDict([("index", "$index")] + format_tt)

    xax, yax = p.axis
    xax.axis_label = 't-SNE coord 1'
    yax.axis_label = 't-SNE coord 2'
    return p


def plot_4(data, *args):
    """Scatter plot of score vs each param
    """
    params = nonconstant_parameters(data)
    scores = np.array([d['mean_test_score'] for d in data])
    order = np.argsort(scores)

    for key in params.keys():
        if params[key].dtype == np.dtype('bool'):
            params[key] = params[key].astype(np.int)
    p_list = []
    for key in params.keys():
        x = params[key][order]
        y = scores[order]
        params = params.loc[order]
        try:
            radius = (np.max(x) - np.min(x)) / 100.0
        except:
            print("error making plot4 for '%s'" % key)
            continue

        p_list.append(build_scatter_tooltip(
            x=x, y=y, radius=radius, add_line=False, tt=params,
            xlabel=key, title='Score vs %s' % key))
    return p_list



from __future__ import print_function, absolute_import, division
"""searchspace.py

This module contains code for specifying the hyperparameter search space.
The search space is specified as a product of bounded intervals. Each dimension
can be either an integer, floating-point or enumeration.

The base measure on the space, e.g. for random sampling (`rvs()`) is a product
of uniform distribution on the bounded intervals. This can, however, be
modified for floating-point dimensions using the `warp` keyword argument in
`add_float`
"""
from collections import namedtuple, Iterable

import numpy as np
from sklearn.utils import check_random_state
try:
    from hyperopt import hp, pyll
except ImportError:
    from .utils import mock_module
    hp = mock_module('hyperopt')
    pyll = mock_module('hyperopt')


class SearchSpace(object):
    def __init__(self):
        self.variables = {}

    @property
    def n_dims(self):
        return len(self.variables)

    def add_jump(self, name, min, max, step, var_type=float):
        """ An integer/float valued dimension bounded between
        range(min, max+step, step). Note that the right endpoint of the interval
        includes `max`. This is a wrapper around the add_enum. It assumes
        that the jump is float but can also use ints. 
        """
        
        choices = np.arange(min, max+step, step, dtype=var_type)
        self.variables[name] = EnumVariable(name, list(choices))

    def add_int(self, name, min, max, warp=None):
        """An integer-valued dimension bounded between `min` <= x <= `max`.
        Note that the right endpoint of the interval includes `max`.

        When `warp` is None, the base measure associated with this dimension
        is a categorical distribution with each weight on each of the integers
        in [min, max]. With `warp == 'log'`, the base measure is a uniform
        distribution on the log of the variable, with bounds at `log(min)` and
        `log(max)`. This is appropriate for variables that are "naturally" in
        log-space. Other `warp` functions are not supported (yet), but may be
        at a later time. Please note that this functionality is not supported
        for `hyperopt_tpe`.
        """
        min, max = map(int, (min, max))
        if max < min:
            raise ValueError('variable %s: max < min error' % name)
        if warp not in (None, 'log'):
            raise ValueError('variable %s: warp=%s is not supported. use '
                             'None or "log",' % (name, warp))
        if min <= 0 and warp == 'log':
            raise ValueError('variable %s: log-warping requires min > 0')

        self.variables[name] = IntVariable(name, min, max, warp)

    def add_float(self, name, min, max, warp=None):
        """A floating point-valued dimension bounded `min` <= x < `max`

        When `warp` is None, the base measure associated with this dimension
        is a uniform distribution on [min, max). With `warp == 'log'`, the
        base measure is a uniform distribution on the log of the variable,
        with bounds at `log(min)` and `log(max)`. This is appropriate for
        variables that are "naturally" in log-space. Other `warp` functions
        are not supported (yet), but may be at a later time.
        """
        min, max = map(float, (min, max))
        if not min < max:
            raise ValueError('variable %s: min >= max error' % name)
        if warp not in (None, 'log'):
            raise ValueError('variable %s: warp=%s is not supported. use '
                             'None or "log",' % (name, warp))
        if min <= 0 and warp == 'log':
            raise ValueError('variable %s: log-warping requires min > 0')

        self.variables[name] = FloatVariable(name, min, max, warp)

    def add_enum(self, name, choices):
        """An enumeration-valued dimension.

        The base measure associated with this dimension is a categorical
        distribution with equal weight on each element in `choices`.
        """
        if not isinstance(choices, Iterable):
            raise ValueError('variable %s: choices must be iterable' % name)
        self.variables[name] = EnumVariable(name, choices)

    def __getitem__(self, name):
        return self.variables[name]

    def __iter__(self):
        return iter(self.variables.values())

    def rvs(self, seed=None):
        random = check_random_state(seed)
        return dict((param.name, param.rvs(random)) for param in self)

    def to_hyperopt(self):
        return dict((v.name, v.to_hyperopt()) for v in self)

    def point_to_gp(self, point_dict):
        return [var.point_to_gp(point_dict[var.name]) for var in self]

    def __repr__(self):
        lines = (['Hyperparameter search space:'] +
                 ['  ' + repr(var) for var in self])
        return '\n'.join(lines)


class IntVariable(namedtuple('IntVariable', ('name', 'min', 'max', 'warp'))):
    # this pattern is a simple memory-efficient way to add some methods to
    # a namedtuple, demonstrated in sklearn.
    # https://github.com/scikit-learn/scikit-learn/blob/a38372998b560d184a195bbd10a16c8f20119aa8/sklearn/grid_search.py#L259-L278

    __slots__ = ()

    def __repr__(self):
        return '{0:<25s}\t(int)   {1:8d} <= x <= {2:d}'.format(
            self.name, self.min, self.max)

    def rvs(self, random):
        # extra +1 here because of the _inclusive_ endpoint
        if self.warp is None:
            return random.randint(self.min, self.max+1)
        elif self.warp == 'log':
            return np.exp(random.uniform(np.log(self.min), np.log(self.max+1)))
        raise ValueError('unknown warp: %s' % self.warp)

    def to_hyperopt(self):
        if self.warp is None:
            return pyll.scope.int(hp.uniform(self.name, self.min, self.max+1))
        raise ValueError('warped integers are not supported for hyperopt')

    def domain_to_gp(self):
        return {'min': 0.0, 'max': 1.0}

    def point_to_gp(self, value):
        if self.warp is None:
            return (value - self.min) / (self.max - self.min)
        elif self.warp == 'log':
            rng = np.log(self.max) - np.log(self.min)
            return (np.log(value) - np.log(self.min)) / rng

        raise ValueError('unknown warp: %s' % self.warp)

    def point_from_gp(self, gpvalue):
        if self.warp is None:
            return int(self.min + (gpvalue * (self.max - self.min)))
        elif self.warp == 'log':
            rng = np.log(self.max) - np.log(self.min)
            outvalue = np.exp(np.log(self.min) + gpvalue * rng)
            return np.clip(outvalue, self.min, self.max).astype(int)
        raise ValueError('unknown warp: %s' % self.warp)


class FloatVariable(namedtuple('FloatVariable',
                               ('name', 'min', 'max', 'warp'))):
    __slots__ = ()

    def __repr__(self):
        return '{0:<25s}\t(float) {1:8f} <= x <  {2:f}'.format(
            self.name, self.min, self.max)

    def rvs(self, random):
        if self.warp is None:
            return random.uniform(self.min, self.max)
        elif self.warp == 'log':
            return np.exp(random.uniform(np.log(self.min), np.log(self.max)))
        raise ValueError('unknown warp: %s' % self.warp)

    def to_hyperopt(self):
        if self.warp is None:
            return hp.uniform(self.name, self.min, self.max)
        elif self.warp == 'log':
            return hp.loguniform(self.name, np.log(self.min), np.log(self.max))
        raise ValueError('unknown warp: %s' % self.warp)

    def domain_to_gp(self):
        return {'min': 0.0, 'max': 1.0}

    def point_to_gp(self, value):
        if self.warp is None:
            return (value - self.min) / (self.max - self.min)
        elif self.warp == 'log':
            rng = np.log(self.max) - np.log(self.min)
            return (np.log(value) - np.log(self.min)) / rng

        raise ValueError('unknown warp: %s' % self.warp)

    def point_from_gp(self, gpvalue):
        if self.warp is None:
            outvalue = self.min + (gpvalue * (self.max - self.min))
        elif self.warp == 'log':
            rng = np.log(self.max) - np.log(self.min)
            outvalue = np.exp(np.log(self.min) + gpvalue * rng)
        else:
            raise ValueError('unknown warp: %s' % self.warp)

        return np.clip(outvalue, self.min, self.max)


class EnumVariable(namedtuple('EnumVariable', ('name', 'choices'))):
    __slots__ = ()

    def __repr__(self):
        c = [str(e) for e in self.choices]
        return '{0:<25s}\t(enum)    choices = ({1:s})'.format(
            self.name, ', '.join(c))

    def rvs(self, random):
        return self.choices[random.randint(len(self.choices))]

    def to_hyperopt(self):
        return hp.choice(self.name, self.choices)

    def domain_to_gp(self):
        return {'min': 0.0, 'max': 1.0}

    def point_to_gp(self, value):
        try:
            index = next(i for i, c in enumerate(self.choices) if c == value)
        except StopIteration:
            raise ValueError('%s not in %s' % (value, self.choices))
        return float(index) / (len(self.choices) - 1)

    def point_from_gp(self, gpvalue):
        return self.choices[int(np.round(gpvalue * (len(self.choices) - 1)))]



from __future__ import print_function, absolute_import, division

import warnings
import pkgutil
import inspect
import importlib

from sklearn.base import BaseEstimator


__all__ = ['msmbuilder', 'import_all_estimators']


def msmbuilder():
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", category=DeprecationWarning)
        import msmbuilder
        from sklearn.pipeline import Pipeline

    scope = import_all_estimators(msmbuilder)
    scope['Pipeline'] = Pipeline
    return scope


def import_all_estimators(pkg):

    def estimator_in_module(mod):
        for name, obj in inspect.getmembers(mod):
            if name.startswith('_'):
                continue
            if inspect.isclass(obj) and issubclass(obj, BaseEstimator):
                yield obj

    result = {}
    for _, modname, ispkg in pkgutil.iter_modules(pkg.__path__):
        c = '%s.%s' % (pkg.__name__, modname)
        if pkg.__name__ + '.tests' in c:
            continue
        try:
            with warnings.catch_warnings():
                warnings.simplefilter("ignore", category=DeprecationWarning)
                mod = importlib.import_module(c)
            if ispkg:
                result.update(import_all_estimators(mod))
            for kls in estimator_in_module(mod):
                if kls.__module__.startswith(pkg.__name__):
                    result[kls.__name__] = kls
        except ImportError as e:
            print('Import Error', c, e)
            continue

    return result



from __future__ import print_function, absolute_import, division

import os
import sqlite3
import shutil
import tempfile

from osprey.trials import make_session, Trial


def test_1():
    cwd = os.path.abspath(os.curdir)
    dirname = tempfile.mkdtemp()
    try:
        os.chdir(dirname)
        session = make_session('sqlite:///db', project_name='abc123')
        session.add(Trial())
        session.commit()

        con = sqlite3.connect('db')
        table_names = con.execute("SELECT name FROM sqlite_master "
                                  "WHERE type='table'").fetchone()
        assert table_names == (u'trials_v3',)

        table_names = con.execute(
            "SELECT project_name FROM trials_v3").fetchone()
        assert table_names == (u'abc123',)

    finally:
        os.chdir(cwd)
        shutil.rmtree(dirname)



from __future__ import print_function, absolute_import, division
from argparse import ArgumentDefaultsHelpFormatter


def func(args, parser):
    # delay import of the rest of the module to improve `osprey -h` performance
    from ..execute_plot import execute
    execute(args, parser)


def configure_parser(sub_parsers):
    help = 'Visualize the collection of jobs in the trials database'
    p = sub_parsers.add_parser('plot', description=help, help=help,
                               formatter_class=ArgumentDefaultsHelpFormatter)
    p.add_argument('config', help='Path to worker config file (yaml)')
    p.add_argument('--filename', default='plot.html')
    g = p.add_mutually_exclusive_group(required=False)
    g.add_argument('--browser', action='store_true', default=True,
                   help='launch browser')
    g.add_argument('--no-browser', action='store_false', dest='browser')

    p.set_defaults(func=func)



from __future__ import print_function, absolute_import, division
"""config.py

This module contains the parser and in-memory representation of the config
osprey job file file. The config file has four major sections:

 - estimator:      the specification for the estimator/model to be fit, an
                   instance of sklearn.base.BaseEstimator.
 - search_space:   the specification of the hyperparameter search space
 - strategy:       strategy for adaptive exploration of hyperparameters.
 - dataset_loader: the specification of the dataset to fit the models with.
 - trials:         as each hyperparameter setting is explored, the results are
                   serialized to a database specified in this section.
 - cv:             specification for cross-validation.
 - scoring:        the score function used in cross-validation. (optional)
"""

import sys
import six
import hashlib
import traceback
import importlib
import contextlib
from os.path import join, isfile, dirname, abspath

import yaml
import sklearn.base
from six.moves import cPickle
from six import iteritems
from six.moves import reduce
from pkg_resources import resource_filename

from .entry_point import load_entry_point
from .utils import dict_merge, in_directory, prepend_syspath, num_samples
from .search_space import SearchSpace
from .strategies import BaseStrategy
from .dataset_loaders import BaseDatasetLoader
from .cross_validators import BaseCVFactory
from .trials import make_session
from .subclass_factory import init_subclass_by_name
from . import eval_scopes


FIELDS = {
    'estimator':       ['pickle', 'eval', 'eval_scope', 'entry_point',
                        'params'],
    'dataset_loader':  ['name', 'params'],
    'trials':          ['uri', 'project_name'],
    'search_space':    dict,
    'strategy':        ['name', 'params'],
    'cv':              (int, dict),
    'scoring':         (str, type(None)),
}


class Config(object):

    def __init__(self, path, verbose=True):
        self.path = path
        self.verbose = verbose
        if not isfile(self.path):
            raise RuntimeError('%s does not exist' % self.path)
        with open(self.path, 'rb') as f:
            config = parse(f)
        self.config = self._merge_defaults(config)
        self._check_fields()

        if self.verbose:
            print('Loading config file:     %s...' % path)

    def _merge_defaults(self, config):
        """The config object loads its values from two sources, with the
        following precedence:

            1. data/default_config.yaml
            2. The config file itself, passed in to this object in the
               constructor as `path`.

        in case of conflict, the config file dominates.
        """
        fn = resource_filename('osprey', join('data', 'default_config.yaml'))
        with open(fn) as f:
            default = parse(f)
        return reduce(dict_merge, [default, config])

    def _check_fields(self):
        for section, submeta in iteritems(self.config):
            if section not in FIELDS:
                raise RuntimeError("unknown section: %s" % section)
            if isinstance(FIELDS[section], type):
                if not isinstance(submeta, FIELDS[section]):
                    raise RuntimeError("%s should be a %s, but is a %s." % (
                        section, FIELDS[section].__name__,
                        type(submeta).__name__))
            elif isinstance(FIELDS[section], tuple):
                if not any(isinstance(submeta, t) for t in FIELDS[section]):
                    raise RuntimeError(
                        "The %s field should be one of %s, not %s" % (
                            section, [e.__name__ for e in FIELDS[section]],
                            type(submeta).__name__))
            else:
                for key in submeta:
                    if key not in FIELDS[section]:
                        raise RuntimeError("in section %r: unknown key %r" % (
                            section, key))

        missing_fields = set(FIELDS.keys()).difference(self.config.keys())
        if len(missing_fields) > 0:
            raise RuntimeError('The following required fields are missing from'
                               'the config file (%s): %s' % (
                                   ', '.join(missing_fields), self.path))

    @classmethod
    def fromdict(cls, config, check_fields=True):
        """Create a Config object from config dict directly."""
        m = super(Config, cls).__new__(cls)
        m.path = '.'
        m.verbose = False
        m.config = m._merge_defaults(config)
        if check_fields:
            m._check_fields()
        return m

    def get_section(self, section):
        """Sections are top-level entries in the config tree"""
        return self.config.get(section, {})

    def get_value(self, field, default=None):
        """Get an entry from within a section, using a '/' delimiter"""
        section, key = field.split('/')
        return self.get_section(section).get(key, default)

    # ----------------------------------------------------------------------- #

    def estimator(self):
        """Get the estimator, an instance of a (subclass of)
        sklearn.base.BaseEstimator

        It can be loaded either from a pickle, from a string using eval(),
        or from an entry point.

        e.g.

        estimator:
            # only one of the following can actually be active in a given
            # config file.
            pickle: path-to-pickle-file.pkl
            eval: "Pipeline([('cluster': KMeans())])"
            entry_point: sklearn.linear_model.LogisticRegression
        """
        evalstring = self.get_value('estimator/eval')
        if evalstring is not None:
            got = self.get_value('estimator/eval_scope')
            if isinstance(got, six.string_types):
                got = [got]
            elif isinstance(got, list):
                pass
            else:
                raise RuntimeError('unexpected type for estimator/eval_scope')

            scope = {}
            for pkg_name in got:
                if pkg_name in eval_scopes.__all__:
                    scope.update(getattr(eval_scopes, pkg_name)())
                else:
                    try:
                        pkg = importlib.import_module(pkg_name)
                    except ImportError as e:
                        raise RuntimeError(str(e))
                    scope.update(eval_scopes.import_all_estimators(pkg))

            try:
                estimator = eval(evalstring, {}, scope)
                if not isinstance(estimator, sklearn.base.BaseEstimator):
                    raise RuntimeError('estimator/pickle must load a '
                                       'sklearn-derived Estimator')
                return estimator
            except:
                print('-'*78, file=sys.stderr)
                print('Error parsing estimator/eval', file=sys.stderr)
                print('-'*78, file=sys.stderr)

                traceback.print_exc(file=sys.stderr)
                print('-'*78, file=sys.stderr)
                sys.exit(1)

        entry_point = self.get_value('estimator/entry_point')
        if entry_point is not None:
            estimator = load_entry_point(entry_point, 'estimator/entry_point')
            if issubclass(estimator, sklearn.base.BaseEstimator):
                estimator = estimator(
                    **self.get_value('estimator/params', default={}))
            if not isinstance(estimator, sklearn.base.BaseEstimator):
                raise RuntimeError('estimator/pickle must load a '
                                   'sklearn-derived Estimator')
            return estimator

        # load estimator from pickle field
        pkl = self.get_value('estimator/pickle')
        if pkl is not None:
            pickl_dir = dirname(abspath(self.path))
            path = join(pickl_dir, pkl)
            if not isfile(path):
                raise RuntimeError('estimator/pickle %s is not a file' % pkl)
            with open(path, 'rb') as f:
                with prepend_syspath(pickl_dir):
                    estimator = cPickle.load(f)
                if not isinstance(estimator, sklearn.base.BaseEstimator):
                    raise RuntimeError('estimator/pickle must load a '
                                       'sklearn-derived Estimator')
                return estimator

        raise RuntimeError('no estimator field')

    def search_space(self):
        ss = self.get_section('search_space')

        searchspace = SearchSpace()
        for param_name, info in iteritems(ss):
            if 'type' not in info:
                raise RuntimeError('search/space/%s does not contain '
                                   'required field "type"' % (param_name))
            type = info.pop('type')
            if type not in ('int', 'float', 'enum', 'jump'):
                raise RuntimeError('search/space/%s type="%s" is not valid. '
                                   'valid types are int, float, enum, and jump'
                                   % param_name)
            try:
                if type == 'int':
                    if sorted(list(info.keys())) not in (
                            ['max', 'min'], ['max', 'min', 'warp']):
                        raise RuntimeError(
                            'search/space/%s type="int" must contain keys '
                            '"min", "max"' % param_name)
                    searchspace.add_int(param_name, **info)
                elif type == 'enum':
                    if sorted(list(info.keys())) != ['choices']:
                        raise RuntimeError(
                            'search/space/%s type="enum" must contain key '
                            '"choices", "type"' % param_name)
                    searchspace.add_enum(param_name, **info)
                elif type == 'float':
                    if sorted(list(info.keys())) not in (
                            ['max', 'min'], ['max', 'min', 'warp']):
                        raise RuntimeError(
                            'search/space/%s type="float" must contain keys '
                            '"min", "max", and optionally "warp"' % param_name)
                    searchspace.add_float(param_name, **info)
                elif type == 'jump':
                    if sorted(list(info.keys())) != ['max', 'min', 'step', 'var_type']:
                        raise RuntimeError(
                            'search/space/%s type="jump" must contain keys '
                            '"min", "max","step" and "var_type"' % param_name)
                    searchspace.add_jump(param_name, **info)
            except ValueError as e:
                # searchspace.add_XXX can throw a ValueError on malformed
                # input (e.g. max < min). re-raising as a runtimerror lets the
                # CLI layer catch it appropriately without an "unexpected
                # error" traceback
                raise RuntimeError(e.message)

        return searchspace

    def strategy(self):
        strategy_name = self.get_value('strategy/name')
        strategy_params = self.get_value('strategy/params', default={})

        strat = init_subclass_by_name(BaseStrategy, strategy_name,
                                      strategy_params)
        return strat

    def dataset(self):
        loader_name = self.get_value('dataset_loader/name')
        loader_params = self.get_value('dataset_loader/params', default={})

        loader = init_subclass_by_name(
            BaseDatasetLoader, loader_name, loader_params)

        with in_directory(dirname(abspath(self.path))):
            X, y = loader.load()

        if num_samples(X) == 0:
            raise RuntimeError('dataset not found.')

        return X, y

    def project_name(self):
        return self.get_value('trials/project_name')

    def trials(self):
        uri = self.get_value('trials/uri')
        project_name = self.project_name()
        if self.verbose:
            print('Loading trials database: %s...' % uri)

        with in_directory(dirname(abspath(self.path))):
            value = make_session(uri, project_name=project_name)
        return value

    @contextlib.contextmanager
    def trialscontext(self):
        session = self.trials()
        yield session
        session.close()

    def scoring(self):
        scoring = self.get_section('scoring')
        assert isinstance(scoring, (str, type(None)))
        return scoring

    def cv(self, X, y=None):
        cv = self.get_section('cv')
        if isinstance(cv, int):
            cv_name = 'kfold'
            cv_params = {'n_folds': cv}
        else:
            cv_name = self.get_value('cv/name')
            cv_params = self.get_value('cv/params', default={})
        return init_subclass_by_name(
            BaseCVFactory, cv_name, cv_params).create(X, y)

    def sha1(self):
        """SHA1 hash of the config file itself."""
        with open(self.path, 'rb') as f:
            return hashlib.sha1(f.read()).hexdigest()


def parse(f):
    res = yaml.load(f)
    if res is None:
        res = {}

    for field, spec in iteritems(FIELDS):
        if field in res:
            if isinstance(spec, type):
                try:
                    res[field] = spec(res[field])
                except ValueError:
                    raise RuntimeError("key %r couldn't be converted to %s" % (
                        field, spec.__name__))
            elif isinstance(spec, tuple):
                if not any(isinstance(res[field], item) for item in spec):
                    raise RuntimeError(
                        "The %s field should be one of %s, not %s" % (
                            field, [e.__name__ for e in spec],
                            res[field].__class__.__name__))

            elif not isinstance(res[field], dict):
                raise RuntimeError("The %s field should be a dict, not %s" % (
                    field, res[field].__class__.__name__))

    return res



from __future__ import print_function

import numpy as np
from .config import Config
from .trials import Trial
from sklearn.pipeline import Pipeline


def execute(args, parser):
    config = Config(args.config, verbose=False)

    session = config.trials()

    items = [curr.to_dict() for curr in session.query(Trial).all()]
    if not items:
        print('No Models Found')
    else:
        mean_test_scores = np.nan_to_num(np.array([i["mean_test_score"]
                                                  for i in items],
                                                  dtype=float))
        c_b_m = items[mean_test_scores.argmax()]
        parameter_dict = c_b_m["parameters"]
        weighted_mean = c_b_m["mean_test_score"]
        raw_test_scores = np.array(c_b_m["test_scores"])
        raw_test_weights = np.array(c_b_m["n_test_samples"])
        weighted_var = np.average((raw_test_scores - weighted_mean)**2, weights=raw_test_weights)
        weighted_std = np.sqrt(weighted_var)
        print('~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~')
        print('Best Current Model = %f +- %f' % (weighted_mean,
                                                 weighted_std))
        print('~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~')
        if isinstance(config.estimator(), Pipeline):
            print('PipelineStep\tParamter \t Value')
            for i in config.estimator().steps:
                print(i[0])
                for param in sorted(parameter_dict.keys()):
                    if str(param).startswith(i[0]):
                        print("\t\t", param.split("__")[1], "\t",
                              parameter_dict[param])
        else:
            print(config.estimator())
            search_space = config.search_space().variables.keys()
            for param in sorted(parameter_dict.keys()):
                if param in search_space:
                    print("\t\t", param, "\t", parameter_dict[param])

    return



from __future__ import print_function, absolute_import, division
import warnings
import numpy as np
import scipy.sparse as sp
import os.path
import sys
import contextlib
from datetime import datetime
from sklearn.pipeline import Pipeline

from .eval_scopes import import_all_estimators


def dict_merge(base, top):
    """Recursively merge two dictionaries, with the elements from `top`
    taking precedence over elements from `top`.

    Returns
    -------
    out : dict
        A new dict, containing the merged records.
    """
    out = dict(top)
    for key in base:
        if key in top:
            if isinstance(base[key], dict) and isinstance(top[key], dict):
                out[key] = dict_merge(base[key], top[key])
        else:
            out[key] = base[key]
    return out


@contextlib.contextmanager
def in_directory(path):
    """Context manager (with statement) that changes the current directory
    during the context.
    """
    curdir = os.path.abspath(os.curdir)
    os.chdir(path)
    yield
    os.chdir(curdir)


@contextlib.contextmanager
def prepend_syspath(path):
    """Contect manager (with statement) that prepends path to sys.path"""
    sys.path.insert(0, path)
    yield
    sys.path.pop(0)


class Unbuffered(object):
    # used to turn off output buffering
    # http://stackoverflow.com/questions/107705/python-output-buffering

    def __init__(self, stream):
        self.stream = stream

    def write(self, data):
        self.stream.write(data)
        self.stream.flush()

    def __getattr__(self, attr):
        return getattr(self.stream, attr)


def format_timedelta(td_object):
    """Format a timedelta object for display to users

    Returns
    -------
    str
    """
    def get_total_seconds(td):
        # timedelta.total_seconds not in py2.6
        return (td.microseconds +
                (td.seconds + td.days * 24 * 3600) * 1e6) / 1e6

    seconds = int(get_total_seconds(td_object))
    periods = [('year',    60*60*24*365),
               ('month',   60*60*24*30),
               ('day',     60*60*24),
               ('hour',    60*60),
               ('minute',  60),
               ('second',  1)]

    strings = []
    for period_name, period_seconds in periods:
        if seconds > period_seconds:
            period_value, seconds = divmod(seconds, period_seconds)
            if period_value == 1:
                strings.append("%s %s" % (period_value, period_name))
            else:
                strings.append("%s %ss" % (period_value, period_name))

    return ", ".join(strings)


def current_pretty_time():
    return datetime.now().strftime("%B %d, %Y %l:%M %p")


def _squeeze_time(t):
    """Remove .1s to the time under Windows: this is the time it take to
    stat files. This is needed to make results similar to timings under
    Unix, for tests
    """
    if sys.platform.startswith('win'):
        return max(0, t - .1)
    else:
        return t


def short_format_time(t):
    t = _squeeze_time(t)
    if t > 60:
        return "%4.1fmin" % (t / 60.)
    else:
        return " %5.1fs" % (t)


def mock_module(name):

    class MockModule(object):
        def __cal__(self, *args, **kwargs):
            raise ImportError('no module named %s' % name)

        def __getattr__(self, *args, **kwargs):
            raise ImportError('no module named %s' % name)

    return MockModule()


def join_quoted(values, quote="'"):
    return ', '.join("%s%s%s" % (quote, e, quote) for e in values)


def expand_path(path, base='.'):
    path = os.path.expanduser(path)
    if not os.path.isabs(path):
        path = os.path.join(base, path)
    return path


def is_msmbuilder_estimator(estimator):
    try:
        import msmbuilder
    except ImportError:
        return False
    msmbuilder_estimators = import_all_estimators(msmbuilder).values()

    out = estimator.__class__ in msmbuilder_estimators
    if isinstance(estimator, Pipeline):
        out = any(step.__class__ in msmbuilder_estimators
                  for name, step in estimator.steps)
    return out


def _assert_all_finite(X):
    """Like assert_all_finite, but only for ndarray."""
    X = np.asanyarray(X)
    # First try an O(n) time, O(1) space solution for the common case that
    # everything is finite; fall back to O(n) space np.isfinite to prevent
    # false positives from overflow in sum method
    if (X.dtype.char in np.typecodes['AllFloat'] and
            not np.isfinite(X.sum()) and not np.isfinite(X).all()):
        raise ValueError("Input contains NaN, infinity"
                         " or a value too large for %r." % X.dtype)


def _warn_if_not_finite(X):
    """UserWarning if array contains non-finite elements"""
    X = np.asanyarray(X)
    # First try an O(n) time, O(1) space solution for the common case that
    # everything is finite; fall back to O(n) space np.isfinite to prevent
    # false positives from overflow in sum method
    if (X.dtype.char in np.typecodes['AllFloat'] and
            not np.isfinite(X.sum()) and not np.isfinite(X).all()):
        warnings.warn("Result contains NaN, infinity"
                      " or a value too large for %r." % X.dtype,
                      category=UserWarning)


def num_samples(x, is_nested=False):
    """Return number of samples in array-like x."""
    if hasattr(x, 'fit'):
        # Don't get num_samples from an ensembles length!
        raise TypeError('Expected sequence or array-like, got '
                        'estimator %s' % x)

    if is_nested:
        return sum(num_samples(xx, is_nested=False) for xx in x)

    if not hasattr(x, '__len__') and not hasattr(x, 'shape'):
        if hasattr(x, '__array__'):
            x = np.asarray(x)
        else:
            raise TypeError("Expected sequence or array-like, got %s" %
                            type(x))
    if hasattr(x, 'shape'):
        if len(x.shape) == 0:
            raise TypeError("Singleton array %r cannot be considered"
                            " a valid collection." % x)
        return x.shape[0]
    else:
        return len(x)


def check_arrays(*arrays, **options):
    """Check that all arrays have consistent first dimensions.

    Checks whether all objects in arrays have the same shape or length.
    By default lists and tuples are converted to numpy arrays.

    It is possible to enforce certain properties, such as dtype, continguity
    and sparse matrix format (if a sparse matrix is passed).

    Converting lists to arrays can be disabled by setting ``allow_lists=True``.
    Lists can then contain arbitrary objects and are not checked for dtype,
    finiteness or anything else but length. Arrays are still checked
    and possibly converted.

    Parameters
    ----------
    *arrays : sequence of arrays or scipy.sparse matrices with same shape[0]
        Python lists or tuples occurring in arrays are converted to 1D numpy
        arrays, unless allow_lists is specified.
    sparse_format : 'csr', 'csc' or 'dense', None by default
        If not None, any scipy.sparse matrix is converted to
        Compressed Sparse Rows or Compressed Sparse Columns representations.
        If 'dense', an error is raised when a sparse array is
        passed.
    copy : boolean, False by default
        If copy is True, ensure that returned arrays are copies of the original
        (if not already converted to another format earlier in the process).
    check_ccontiguous : boolean, False by default
        Check that the arrays are C contiguous
    dtype : a numpy dtype instance, None by default
        Enforce a specific dtype.
    warn_nans : boolean, False by default
        Prints warning if nans in the arrays
        Disables allow_nans
    replace_nans : boolean, False by default
        Replace nans in the arrays with zeros
    allow_lists : bool
        Allow lists of arbitrary objects as input, just check their length.
        Disables
    allow_nans : boolean, False by default
        Allows nans in the arrays
    allow_nd : boolean, False by default
        Allows arrays of more than 2 dimensions.
    """
    sparse_format = options.pop('sparse_format', None)
    if sparse_format not in (None, 'csr', 'csc', 'dense'):
        raise ValueError('Unexpected sparse format: %r' % sparse_format)
    copy = options.pop('copy', False)
    check_ccontiguous = options.pop('check_ccontiguous', False)
    dtype = options.pop('dtype', None)
    warn_nans = options.pop('warn_nans', False)
    replace_nans = options.pop('replace_nans', False)
    allow_lists = options.pop('allow_lists', False)
    allow_nans = options.pop('allow_nans', False)
    allow_nd = options.pop('allow_nd', False)

    if options:
        raise TypeError("Unexpected keyword arguments: %r" % options.keys())

    if len(arrays) == 0:
        return None

    n_samples = num_samples(arrays[0])

    checked_arrays = []
    for array in arrays:
        array_orig = array
        if array is None:
            # special case: ignore optional y=None kwarg pattern
            checked_arrays.append(array)
            continue
        size = num_samples(array)

        if size != n_samples:
            raise ValueError("Found array with dim %d. Expected %d"
                             % (size, n_samples))

        if not allow_lists or hasattr(array, "shape"):
            if sp.issparse(array):
                if sparse_format == 'csr':
                    array = array.tocsr()
                elif sparse_format == 'csc':
                    array = array.tocsc()
                elif sparse_format == 'dense':
                    raise TypeError('A sparse matrix was passed, but dense '
                                    'data is required. Use X.toarray() to '
                                    'convert to a dense numpy array.')
                if check_ccontiguous:
                    array.data = np.ascontiguousarray(array.data, dtype=dtype)
                elif hasattr(array, 'data'):
                    array.data = np.asarray(array.data, dtype=dtype)
                elif array.dtype != dtype:
                    array = array.astype(dtype)
                if not allow_nans:
                    if hasattr(array, 'data'):
                        _assert_all_finite(array.data)
                    else:
                        _assert_all_finite(array.values())
            else:
                if check_ccontiguous:
                    array = np.ascontiguousarray(array, dtype=dtype)
                else:
                    array = np.asarray(array, dtype=dtype)
                if warn_nans:
                    allow_nans = True
                    _warn_if_not_finite(array)
                if replace_nans:
                    array = np.nan_to_num(array)
                if not allow_nans:
                    _assert_all_finite(array)

            if not allow_nd and array.ndim >= 3:
                raise ValueError("Found array with dim %d. Expected <= 2" %
                                 array.ndim)

        if copy and array is array_orig:
            array = array.copy()
        checked_arrays.append(array)

    return checked_arrays



from __future__ import print_function, absolute_import, division

import sys
from six import iteritems
import numpy as np
from numpy.testing.decorators import skipif

from osprey.search_space import SearchSpace
from osprey.search_space import IntVariable, EnumVariable, FloatVariable
from osprey.strategies import RandomSearch, HyperoptTPE, GP, GridSearch

try:
    from hyperopt import hp, fmin, tpe, Trials
except:
    pass


def test_random():
    searchspace = SearchSpace()
    searchspace.add_float('x', -10, 10)
    random = np.random.RandomState(0)
    RandomSearch(seed=random).suggest([], searchspace)


def test_grid():
    searchspace = SearchSpace()
    searchspace.add_enum('x', [1, 2])
    searchspace.add_jump('y', min=3, max=4, step=1)
    grid_search = GridSearch()
    suggestions = [grid_search.suggest([], searchspace) for _ in range(4)]
    suggestions = [(s['x'], s['y']) for s in suggestions]
    assert suggestions == [(1, 3), (1, 4), (2, 3), (2, 4)], "Didn't examine whole space correctly"


def hyperopt_x2_iterates(n_iters=100):
    iterates = []
    trials = Trials()
    random = np.random.RandomState(0)

    def fn(params):
        iterates.append(params['x'])
        return params['x']**2

    for i in range(n_iters):
        fmin(fn=fn, algo=tpe.suggest, max_evals=i+1, trials=trials,
             space={'x': hp.uniform('x', -10, 10)},
             **HyperoptTPE._hyperopt_fmin_random_kwarg(random))

    return np.array(iterates)


def our_x2_iterates(n_iters=100):
    history = []
    searchspace = SearchSpace()
    searchspace.add_float('x', -10, 10)
    random = np.random.RandomState(0)

    # note the switch of sign, because _our_ function hyperopt_tpe is
    # a maximizer, not a minimizer
    def fn(params):
        return -params['x']**2

    for i in range(n_iters):
        params = HyperoptTPE(seed=random).suggest(history, searchspace)
        history.append((params, fn(params), 'SUCCEEDED'))

    return np.array([h[0]['x'] for h in history])


@skipif('hyperopt.fmin' not in sys.modules, 'this test requires hyperopt')
def test_1():
    ours = our_x2_iterates(25)
    ref = hyperopt_x2_iterates(25)

    np.testing.assert_array_equal(ref, ours)


@skipif('GPy' not in sys.modules, 'this test requires hyperopt')
def test_gp():
    searchspace = SearchSpace()
    searchspace.add_float('x', -10, 10)
    searchspace.add_float('y', 1, 10, warp='log')
    searchspace.add_int('z', -10, 10)
    searchspace.add_enum('w', ['opt1', 'opt2'])

    history = [(searchspace.rvs(), np.random.random(), 'SUCCEEDED')
               for _ in range(4)]
    params = GP().suggest(history, searchspace)
    for k, v in iteritems(params):
        assert k in searchspace.variables
        if isinstance(searchspace[k], EnumVariable):
            assert v in searchspace[k].choices
        elif isinstance(searchspace[k], FloatVariable):
            assert searchspace[k].min <= v <= searchspace[k].max
        elif isinstance(searchspace[k], IntVariable):
            assert searchspace[k].min <= v <= searchspace[k].max
        else:
            assert False



from __future__ import print_function, absolute_import, division
import json
import shutil
import sys
import tempfile
import unittest
from argparse import Namespace
import numpy as np
from numpy.testing.decorators import skipif

try:
    import pylearn2
    del pylearn2  # for flake8, since not used in tests
except ImportError:
    pass

from osprey import execute_worker, execute_dump
from osprey.plugins.plugin_pylearn2 import Pylearn2DatasetLoader


class TestPluginPylearn2(unittest.TestCase):
    """
    Test plugin_pylearn2.
    """
    def setUp(self):
        self.temp_dir = tempfile.mkdtemp()

    def tearDown(self):
        shutil.rmtree(self.temp_dir)

    def run_osprey(self, config):
        """
        Run osprey-worker.

        Parameters
        ----------
        config : str
            Configuration string.
        """
        fh, filename = tempfile.mkstemp(dir=self.temp_dir)
        with open(filename, 'wb') as f:
            f.write(config)
        args = Namespace(config=filename, n_iters=1, output='json')
        execute_worker.execute(args, None)
        dump = json.loads(execute_dump.execute(args, None))
        assert len(dump) == 1
        assert dump[0]['status'] == 'SUCCEEDED', dump[0]['status']

    @skipif('pylearn2' not in sys.modules, 'this test requires pylearn2')
    def test_pylearn2_classifier(self):
        config = """
estimator:
  entry_point: osprey.plugins.plugin_pylearn2.Pylearn2Classifier
  params:
    yaml_string: "!obj:pylearn2.train.Train {
  dataset: null,
  model: !obj:pylearn2.models.mlp.MLP {
    nvis: 4,
    layers: [
      !obj:pylearn2.models.mlp.Sigmoid {
        layer_name: h0,
        dim: $dim,
        irange: 0.05,
      },
      !obj:pylearn2.models.mlp.Softmax {
        layer_name: y,
        n_classes: 3,
        irange: 0.,
      },
    ],
  },
  algorithm: !obj:pylearn2.training_algorithms.bgd.BGD {
    batch_size: 10,
    termination_criterion: !obj:pylearn2.termination_criteria.EpochCounter {
      max_epochs: 1,
    },
  },
}"

dataset_loader:
  name: sklearn_dataset
  params:
    method: load_iris

strategy:
  name: random

search_space:
  dim:
    min: 1
    max: 10
    type: int

cv:
  name: stratifiedkfold

trials:
  uri: sqlite:///test.db
"""
        self.run_osprey(config)

    @skipif('pylearn2' not in sys.modules, 'this test requires pylearn2')
    def test_pylearn2_regressor(self):
        config = """
estimator:
  entry_point: osprey.plugins.plugin_pylearn2.Pylearn2Regressor
  params:
    yaml_string: "!obj:pylearn2.train.Train {
  dataset: null,
  model: !obj:pylearn2.models.mlp.MLP {
    nvis: 13,
    layers: [
      !obj:pylearn2.models.mlp.Sigmoid {
        layer_name: h0,
        dim: $dim,
        irange: 0.05,
      },
      !obj:pylearn2.models.mlp.Linear {
        layer_name: y,
        dim: 1,
        irange: 0.,
      },
    ],
  },
  algorithm: !obj:pylearn2.training_algorithms.bgd.BGD {
    batch_size: 10,
    termination_criterion: !obj:pylearn2.termination_criteria.EpochCounter {
      max_epochs: 1,
    },
  },
}"

dataset_loader:
  name: sklearn_dataset
  params:
    method: load_boston

strategy:
  name: random

search_space:
  dim:
    min: 1
    max: 10
    type: int

cv: 3

trials:
  uri: sqlite:///test.db
"""
        self.run_osprey(config)

    @skipif('pylearn2' not in sys.modules, 'this test requires pylearn2')
    def test_pylearn2_autoencoder(self):
        config = """
estimator:
  entry_point: osprey.plugins.plugin_pylearn2.Pylearn2Autoencoder
  params:
    yaml_string: "!obj:pylearn2.train.Train {
  dataset: null,
  model: !obj:pylearn2.models.autoencoder.Autoencoder {
    act_dec: null,
    act_enc: sigmoid,
    irange: 0.05,
    nvis: 13,
    nhid: $dim,
  },
  algorithm: !obj:pylearn2.training_algorithms.bgd.BGD {
    batch_size: 10,
    cost: !obj:pylearn2.costs.autoencoder.MeanSquaredReconstructionError {},
    termination_criterion: !obj:pylearn2.termination_criteria.EpochCounter {
      max_epochs: 1,
    },
  },
}"

dataset_loader:
  name: sklearn_dataset
  params:
    method: load_boston

strategy:
  name: random

search_space:
  dim:
    min: 1
    max: 10
    type: int

cv: 3

trials:
  uri: sqlite:///test.db
"""
        self.run_osprey(config)

    @skipif('pylearn2' not in sys.modules, 'this test requires pylearn2')
    def test_pylearn2_dataset_loader(self):
        yaml_string = """
!obj:pylearn2.testing.datasets.random_dense_design_matrix {
  rng: !obj:numpy.random.RandomState { seed: 1 },
  num_examples: 10,
  dim: 5,
  num_classes: 3,
}
"""
        dataset_loader = Pylearn2DatasetLoader(yaml_string, one_hot=False)
        X, y = dataset_loader.load()
        assert X.shape == (10, 5)
        assert y.shape == (10,)
        assert np.unique(y).size == 3

    @skipif('pylearn2' not in sys.modules, 'this test requires pylearn2')
    def test_pylearn2_dataset_loader_one_hot(self):
        yaml_string = """
!obj:pylearn2.testing.datasets.random_one_hot_dense_design_matrix {
  rng: !obj:numpy.random.RandomState { seed: 1 },
  num_examples: 10,
  dim: 5,
  num_classes: 3,
}
"""
        dataset_loader = Pylearn2DatasetLoader(yaml_string, one_hot=True)
        X, y = dataset_loader.load()
        assert X.shape == (10, 5)
        assert y.shape == (10,)
        assert np.unique(y).size == 3



from __future__ import print_function, absolute_import, division

import numpy as np
from nose.plugins.skip import SkipTest
from six import iteritems
from sklearn.datasets import make_regression
from sklearn.linear_model import Lasso
from sklearn.grid_search import GridSearchCV

from osprey.fit_estimator import fit_and_score_estimator


def test_1():
    X, y = make_regression(n_features=10)

    lasso = Lasso()
    params = {'alpha': 2}
    cv = 6
    out = fit_and_score_estimator(lasso, params, cv=cv, X=X, y=y, verbose=0)

    param_grid = dict((k, [v]) for k, v in iteritems(params))
    g = GridSearchCV(estimator=lasso, param_grid=param_grid, cv=cv)
    g.fit(X, y)

    np.testing.assert_almost_equal(
        out['mean_test_score'], g.grid_scores_[0].mean_validation_score)

    assert np.all(out['test_scores'] == g.grid_scores_[0].cv_validation_scores)


def test_2():
    try:
        from msmbuilder.msm import MarkovStateModel
    except ImportError as e:
        raise SkipTest(e)

    X = [np.random.randint(2, size=10), np.random.randint(2, size=11)]
    out = fit_and_score_estimator(
        MarkovStateModel(), {'verbose': False}, cv=2, X=X, y=None, verbose=0)
    np.testing.assert_array_equal(out['n_train_samples'], [11, 10])
    np.testing.assert_array_equal(out['n_test_samples'], [10, 11])



from __future__ import print_function, absolute_import, division
import os
import shutil
import tempfile

import numpy as np
import sklearn.datasets
from sklearn.externals.joblib import dump

from osprey.dataset_loaders import (FilenameDatasetLoader, JoblibDatasetLoader,
                                    HDF5DatasetLoader, MDTrajDatasetLoader,
                                    MSMBuilderDatasetLoader,
                                    NumpyDatasetLoader, SklearnDatasetLoader)


def test_FilenameDatasetLoader_1():
    cwd = os.path.abspath(os.curdir)
    dirname = tempfile.mkdtemp()
    try:
        os.chdir(dirname)

        open('filename-1', 'w').close()
        open('filename-2', 'w').close()

        assert FilenameDatasetLoader.short_name == 'filename'
        loader = FilenameDatasetLoader('filename-*')
        X, y = loader.load()

        X_ref = list(map(os.path.abspath, ['filename-1', 'filename-2']))
        assert sorted(X) == X_ref, X
        assert y is None, y

    finally:
        os.chdir(cwd)
        shutil.rmtree(dirname)


def test_JoblibDatasetLoader_1():
    assert JoblibDatasetLoader.short_name == 'joblib'

    cwd = os.path.abspath(os.curdir)
    dirname = tempfile.mkdtemp()
    try:
        os.chdir(dirname)

        # one file
        dump(np.zeros((10, 2)), 'f1.pkl')
        loader = JoblibDatasetLoader('f1.pkl')
        X, y = loader.load()
        assert np.all(X == np.zeros((10, 2)))
        assert y is None

        # two files
        dump(np.ones((10, 2)), 'f2.pkl')
        loader = JoblibDatasetLoader('f*.pkl')
        X, y = loader.load()
        assert isinstance(X, list)
        assert np.all(X[0] == np.zeros((10, 2)))
        assert np.all(X[1] == np.ones((10, 2)))
        assert y is None

        # one file, with x and y
        dump({'foo': 'baz', 'bar': 'qux'}, 'foobar.pkl')
        loader = JoblibDatasetLoader('foobar.pkl', x_name='foo', y_name='bar')
        X, y = loader.load()
        assert X == 'baz', X
        assert y == 'qux', y

    finally:
        os.chdir(cwd)
        shutil.rmtree(dirname)


def test_HDF5DatasetLoader_1():
    from mdtraj import io

    assert HDF5DatasetLoader.short_name == 'hdf5'

    cwd = os.path.abspath(os.curdir)
    dirname = tempfile.mkdtemp()
    try:
        os.chdir(dirname)

        # one file
        io.saveh('f1.h5', **{'test': np.zeros((10, 2))})
        loader = HDF5DatasetLoader('f1.h5')
        X, y = loader.load()
        assert np.all(X == np.zeros((10, 2)))
        assert y[0] == 0

        # two files
        io.saveh('f2.h5', **{'test': np.ones((10, 2))})
        loader = HDF5DatasetLoader('f*.h5')
        X, y = loader.load()
        assert isinstance(X, list)
        assert np.all(X[0] == np.zeros((10, 2)))
        assert np.all(X[1] == np.ones((10, 2)))
        assert y[1] == 1

    finally:
        os.chdir(cwd)
        shutil.rmtree(dirname)


def test_MDTrajDatasetLoader_1():
    from mdtraj.testing import get_fn

    loader = MDTrajDatasetLoader(get_fn('legacy_msmbuilder_trj0.lh5'))
    X, y = loader.load()
    assert X[0].n_frames == 501
    assert y is None


def test_MSMBuilderDatasetLoader_1():
    from msmbuilder.dataset import dataset

    path = tempfile.mkdtemp()
    shutil.rmtree(path)
    try:
        x = np.random.randn(10, 2)
        ds = dataset(path, 'w', 'dir-npy')
        ds[0] = x

        loader = MSMBuilderDatasetLoader(path, fmt='dir-npy')
        X, y = loader.load()

        assert np.all(X[0] == x)
        assert y is None

    finally:
        shutil.rmtree(path)


def test_NumpyDatasetLoader_1():
        cwd = os.path.abspath(os.curdir)
        dirname = tempfile.mkdtemp()
        try:
            os.chdir(dirname)

            x = np.random.randn(10, 2)
            np.save('f1.npy', x)

            loader = NumpyDatasetLoader('f1.npy')
            X, y = loader.load()

            assert np.all(X[0] == x)
            assert y is None

        finally:
            os.chdir(cwd)
            shutil.rmtree(dirname)


def test_SklearnDatasetLoader_1():
    assert SklearnDatasetLoader.short_name == 'sklearn_dataset'
    X, y = SklearnDatasetLoader('load_iris').load()
    iris = sklearn.datasets.load_iris()
    assert np.all(X == iris['data'])
    assert np.all(y == iris['target'])



from __future__ import print_function, absolute_import, division

import inspect
import itertools
import collections
from six import string_types
from .utils import join_quoted


def init_subclass_by_name(baseclass, short_name, params):
    """
    Find the subclass, `kls` of baseclass with class attribute `short_name`
    that matches the supplied `short_name`, and then instantiate and return
    that class with:

        return kls(**params)

    This function also tries its best to catch any possible TypeErrors due
    to binding of the arguments, and rethrows them as nicely formatted
    RuntimeErrors that are suitable for showing to users.
    """
    sc = baseclass.__subclasses__()
    for kls in sc:
        if kls.short_name == short_name or \
           (_is_collection(kls.short_name) and short_name in kls.short_name):
            try:
                return kls(**params)
            except TypeError as e:
                spec = inspect.getargspec(kls.__init__)

                # try to give nice errors to the user if the params failed
                # to bind
                if 'unexpected' in str(e):
                    avail = join_quoted(spec.args[1:])
                    raise RuntimeError(
                        "%s's %s. Available params for this subclass are: %s."
                        % (short_name, str(e), avail))
                elif 'takes exactly' in str(e):
                    required = join_quoted(spec.args[1:-len(spec.defaults)])
                    raise RuntimeError(
                        "%s's %s. Required params for this subclass are %s."
                        % (short_name, str(e), required))
                elif 'takes at least' in str(e):
                    required = join_quoted(spec.args[1:-len(spec.defaults)])
                    optional = join_quoted(spec.args[-len(spec.defaults):])
                    raise RuntimeError(
                        "%s's %s. Required params for this subclass are: %s. "
                        "Optional params are: %s" % (
                            short_name, str(e), required, optional))
                # :(
                raise

    chain = itertools.chain.from_iterable(
        e.short_name if _is_collection(e.short_name) else [e.short_name]
        for e in sc)
    avail_names = ', '.join(str(n) for n in chain)
    raise ValueError('"%s" is not a recognized subclass. available names '
                     'are: %s' % (short_name, avail_names))


def _is_collection(obj):
    if isinstance(obj, string_types):
        return False
    return isinstance(obj, collections.Iterable)



from __future__ import print_function, absolute_import, division
import os
import sys
import json
import shutil
import subprocess
import tempfile
from distutils.spawn import find_executable
from numpy.testing.decorators import skipif

try:
    __import__('msmbuilder')
    HAVE_MSMBUILDER = True
except:
    HAVE_MSMBUILDER = False

OSPREY_BIN = find_executable('osprey')


@skipif(not HAVE_MSMBUILDER, 'this test requires MSMBuilder')
def test_1():
    from msmbuilder.example_datasets import FsPeptide
    assert OSPREY_BIN is not None
    cwd = os.path.abspath(os.curdir)
    dirname = tempfile.mkdtemp()
    FsPeptide(dirname).get()

    try:
        os.chdir(dirname)
        subprocess.check_call([OSPREY_BIN, 'skeleton', '-t', 'msmbuilder',
                              '-f', 'config.yaml'])
        subprocess.check_call([OSPREY_BIN, 'worker', 'config.yaml', '-n', '1'])
        assert os.path.exists('osprey-trials.db')

        yield _test_dump_1

        yield _test_plot_1

    finally:
        os.chdir(cwd)
        shutil.rmtree(dirname)


def test_2():
    assert OSPREY_BIN is not None
    cwd = os.path.abspath(os.curdir)
    dirname = tempfile.mkdtemp()

    try:
        os.chdir(dirname)
        subprocess.check_call([OSPREY_BIN, 'skeleton', '-t', 'sklearn',
                              '-f', 'config.yaml'])
        subprocess.check_call([OSPREY_BIN, 'worker', 'config.yaml', '-n', '1'])
        assert os.path.exists('osprey-trials.db')

        subprocess.check_call([OSPREY_BIN, 'current_best', 'config.yaml'])

        yield _test_dump_1

        yield _test_plot_1

    finally:
        os.chdir(cwd)
        shutil.rmtree(dirname)


def _test_dump_1():
    out = subprocess.check_output(
        [OSPREY_BIN, 'dump', 'config.yaml', '-o', 'json'])
    if sys.version_info >= (3, 0):
        out = out.decode()
    json.loads(out)


def _test_plot_1():
    _ = subprocess.check_output(
        [OSPREY_BIN, 'plot', 'config.yaml', '--no-browser'])
    if not os.path.isfile('./plot.html'):
        raise ValueError('Plot not created')



from __future__ import print_function, absolute_import, division

import glob
import os
import numpy as np

from .utils import expand_path, num_samples


class BaseDatasetLoader(object):
    short_name = None

    def load(self):
        raise NotImplementedError('should be implemented in subclass')


class MSMBuilderDatasetLoader(BaseDatasetLoader):
    short_name = 'msmbuilder'

    def __init__(self, path, fmt=None, verbose=False):
        self.path = path
        self.fmt = fmt
        self.verbose = verbose

    def load(self):
        from msmbuilder.dataset import dataset
        ds = dataset(self.path, mode='r', fmt=self.fmt, verbose=self.verbose)
        print('Dataset provenance:\n')
        print(ds.provenance)
        return ds, None


class NumpyDatasetLoader(BaseDatasetLoader):
    short_name = 'numpy'

    def __init__(self, filenames):
        self.filenames = filenames

    def load(self):
        filenames = sorted(glob.glob(expand_path(self.filenames)))
        if len(filenames) == 0:
            raise RuntimeError('no filenames matched by pattern: %s' %
                               self.filenames)
        ds = [np.load(f) for f in filenames]
        return ds, None


class HDF5DatasetLoader(BaseDatasetLoader):
    short_name = 'hdf5'

    def __init__(self, filenames, stride=1):
        self.filenames = filenames
        self.stride = stride

    def transform(self, arr):
        return arr[::self.stride, :]

    def load(self):
        from mdtraj import io
        X = []
        y = []
        filenames = sorted(glob.glob(expand_path(self.filenames)))
        for i, fn in enumerate(filenames):
            dataset = io.loadh(fn)
            for key in dataset.iterkeys():
                X.append(self.transform(dataset[key]))
                y.append(i)
        return X, y


class MDTrajDatasetLoader(BaseDatasetLoader):
    short_name = 'mdtraj'

    def __init__(self, trajectories, topology=None, stride=1, verbose=False):
        self.trajectories = trajectories
        self.topology = topology
        self.stride = stride
        self.verbose = verbose

    def load(self):
        import mdtraj

        filenames = sorted(glob.glob(expand_path(self.trajectories)))
        if len(filenames) == 0:
            raise RuntimeError('no filenames matched by pattern: %s' %
                               self.trajectories)

        top = self.topology
        kwargs = {}
        if top is not None:
            top = expand_path(self.topology)
            kwargs = {'top': top}

        X = []
        y = None

        for fn in filenames:
            if self.verbose:
                print('[mdtraj] loading %s' % fn)
            X.append(mdtraj.load(fn, stride=self.stride, **kwargs))

        return X, y


class FilenameDatasetLoader(BaseDatasetLoader):
    """Just pass a bunch of filenames to the first step of the pipeline

    The pipeline will do the loading.
    """
    short_name = 'filename'

    def __init__(self, trajectories, abs_path=True):
        self.traj_glob = trajectories
        self.abs_path = abs_path

    def load(self):
        filenames = sorted(glob.glob(expand_path(self.traj_glob)))
        if len(filenames) == 0:
            raise RuntimeError('no filenames matched by pattern: %s' %
                               self.traj_glob)

        if self.abs_path:
            filenames = [os.path.abspath(fn) for fn in filenames]
        return filenames, None


class JoblibDatasetLoader(BaseDatasetLoader):
    short_name = 'joblib'

    def __init__(self, filenames, x_name=None, y_name=None,
                 system_joblib=False):
        self.filenames = filenames
        self.x_name = x_name
        self.y_name = y_name
        self.system_joblib = system_joblib

    def load(self):
        if self.system_joblib:
            import joblib
        else:
            from sklearn.externals import joblib

        X, y = [], []

        filenames = sorted(glob.glob(expand_path(self.filenames)))
        if len(filenames) == 0:
            raise RuntimeError('no filenames matched by pattern: %s' %
                               self.filenames)

        for fn in filenames:
            obj = joblib.load(fn)
            if isinstance(obj, (list, np.ndarray)):
                X.append(obj)
            else:
                X.append(obj[self.x_name])
                y.append(obj[self.y_name])

        if num_samples(X) == 1:
            X = X[0]
        if len(y) == 1:
            y = y[0]
        elif len(y) == 0:
            y = None

        return X, y


class SklearnDatasetLoader(BaseDatasetLoader):
    short_name = 'sklearn_dataset'

    def __init__(self, method, x_name='data', y_name='target', **kwargs):
        self.method = method
        self.x_name = x_name
        self.y_name = y_name
        self.kwargs = kwargs

    def load(self):
        import sklearn.datasets

        try:
            loader = getattr(sklearn.datasets, self.method)
        except AttributeError:
            raise RuntimeError('no %s in sklearn.datasets' % self.method)

        bunch = loader(**self.kwargs)

        X = bunch[self.x_name]
        y = bunch[self.y_name]

        return X, y



from __future__ import print_function, absolute_import, division
import os
import shutil
import subprocess
import tempfile
import yaml
from distutils.spawn import find_executable

from osprey.config import Config

OSPREY_BIN = find_executable('osprey')


def test_1():
    assert OSPREY_BIN is not None
    cwd = os.path.abspath(os.curdir)
    dirname = tempfile.mkdtemp()

    try:
        os.chdir(dirname)
        subprocess.check_call([OSPREY_BIN, 'skeleton', '-t', 'msmbuilder',
                              '-f', 'config.yaml'])
        assert os.path.exists('config.yaml')
        with open('config.yaml', 'rb') as f:
            yaml.load(f)
        Config('config.yaml')

    finally:
        os.chdir(cwd)
        shutil.rmtree(dirname)



from __future__ import print_function, absolute_import, division

import sys
import numpy as np
import scipy.stats
from six.moves import xrange
from numpy.testing.decorators import skipif
try:
    from hyperopt import pyll
except ImportError:
    pass

from osprey.search_space import SearchSpace
from osprey.search_space import IntVariable, FloatVariable, EnumVariable


def test_1():
    s = SearchSpace()
    s.add_int('a', 1, 2)
    s.add_float('b', 2, 3)
    s.add_enum('c', ['a', 'b', 'c'])

    assert s.n_dims == 3

    assert s['a'].min == 1
    assert s['a'].max == 2
    assert s['a'].name == 'a'

    assert s['b'].min == 2
    assert s['b'].max == 3
    assert s['b'].name == 'b'

    assert s['c'].choices == ['a', 'b', 'c']
    assert s['c'].name == 'c'


def _run_chi2_test(values, bin_edges):
    n_samples = len(values)
    n_bins = len(bin_edges) - 1
    counts, bin_edges = np.histogram(values, bin_edges)
    p = scipy.stats.chisquare(counts, f_exp=[n_samples/n_bins] * n_bins)[1]
    if p < 0.001:
        raise ValueError('distribution not being sampled correctly, p=%f' % p)


def test_2_1():
    s = SearchSpace()
    s.add_int('a', 0, 3)

    values = [s.rvs()['a'] for _ in xrange(100)]
    _run_chi2_test(values, bin_edges=range(5))


@skipif('hyperopt.pyll' not in sys.modules, 'this test requires hyperopt')
def test_2_2():
    s = SearchSpace()
    s.add_int('a', 0, 3)
    values = [pyll.stochastic.sample(s['a'].to_hyperopt())
              for _ in xrange(200)]
    _run_chi2_test(values, bin_edges=range(5))


def test_3_1():
    s = SearchSpace()
    s.add_float('b', -2, 2)

    values = [s.rvs()['b'] for _ in range(100)]
    assert all(-2 < v < 2 for v in values)
    _run_chi2_test(values, bin_edges=np.linspace(-2, 2, 10))


@skipif('hyperopt.pyll' not in sys.modules, 'this test requires hyperopt')
def test_3_2():
    s = SearchSpace()
    s.add_float('b', -2, 2)

    values = [pyll.stochastic.sample(s['b'].to_hyperopt())
              for _ in xrange(100)]
    assert all(-2 < v < 2 for v in values)
    _run_chi2_test(values, bin_edges=np.linspace(-2, 2, 10))


def test_4_1():
    s = SearchSpace()
    s.add_enum('c', [True, False])

    values = [s.rvs()['c'] for _ in range(100)]
    assert all(v in [True, False] for v in values)
    _run_chi2_test(np.array(values, dtype=int), bin_edges=range(3))


@skipif('hyperopt.pyll' not in sys.modules, 'this test requires hyperopt')
def test_4_2():
    s = SearchSpace()
    s.add_enum('c', [True, False])

    values = [pyll.stochastic.sample(s['c'].to_hyperopt())
              for _ in xrange(100)]
    assert all(v in [True, False] for v in values)
    _run_chi2_test(np.array(values, dtype=int), bin_edges=range(3))


def test_5_1():
    s = SearchSpace()
    s.add_float('a', 1e-5, 1, warp='log')

    n_bins = 10
    n_samples = 1000

    bin_edges = np.logspace(np.log10(s['a'].min), np.log10(s['a'].max),
                            num=n_bins+1)
    values = [s.rvs()['a'] for _ in xrange(n_samples)]

    _run_chi2_test(values, bin_edges)


@skipif('hyperopt.pyll' not in sys.modules, 'this test requires hyperopt')
def test_5_2():
    s = SearchSpace()
    s.add_float('a', 1e-5, 1, warp='log')

    bin_edges = np.logspace(np.log10(s['a'].min), np.log10(s['a'].max), num=5)
    values = [pyll.stochastic.sample(s['a'].to_hyperopt())
              for _ in xrange(100)]

    _run_chi2_test(values, bin_edges)


def test_gp_1():
    v = IntVariable('name', 1, 10, None)
    assert 2 == v.point_from_gp(v.point_to_gp(2))
    assert 1 == v.point_from_gp(v.point_to_gp(1))
    assert 10 == v.point_from_gp(v.point_to_gp(10))

    assert v.point_to_gp(1) == 0
    assert v.point_to_gp(10) == 1


def test_gp_2():
    v = FloatVariable('name', 1, 10, None)
    assert (2-1) / (10-1) == v.point_to_gp(2)
    assert 0.0 == v.point_to_gp(1)
    assert 1.0 == v.point_to_gp(10)
    assert 2.0 == v.point_from_gp(v.point_to_gp(2))
    assert 1.0 == v.point_from_gp(v.point_to_gp(1))
    assert 1.0 == v.point_from_gp(v.point_to_gp(0.9))
    assert 10.0 == v.point_from_gp(v.point_to_gp(10))
    assert 10.0 == v.point_from_gp(v.point_to_gp(10.1))


def test_gp_3():
    v = FloatVariable('name', 1, 10, 'log')
    assert 2.0 == v.point_from_gp(v.point_to_gp(2))
    assert 1.0 == v.point_from_gp(v.point_to_gp(1))
    assert 1.0 == v.point_from_gp(v.point_to_gp(0.9))
    assert 10.0 == v.point_from_gp(v.point_to_gp(10))
    assert 10.0 == v.point_from_gp(v.point_to_gp(10.1))


def test_gp_4():
    v = EnumVariable('name', ['a', 'b', 'c'])
    assert 'a' == v.point_from_gp(v.point_to_gp('a'))
    assert 'b' == v.point_from_gp(v.point_to_gp('b'))
    assert 'c' == v.point_from_gp(v.point_to_gp('c'))
    assert 0 == v.point_to_gp('a')
    assert 0.5 == v.point_to_gp('b')
    assert 1 == v.point_to_gp('c')



from __future__ import print_function, absolute_import, division
from osprey.entry_point import load_entry_point
from nose.tools import assert_raises


def test_1():
    from sklearn.cluster import KMeans
    assert load_entry_point('sklearn.cluster.KMeans') is KMeans


def test_2():
    from numpy.random import randint
    assert load_entry_point('numpy.random.randint') is randint


def test_3():
    assert load_entry_point('osprey.entry_point.load_entry_point') \
        is load_entry_point


def test_4():
    assert_raises(RuntimeError,
                  lambda: load_entry_point('sklearn'))
    assert_raises(RuntimeError,
                  lambda: load_entry_point('sklearn.sdsdfjhgdsf'))



from __future__ import print_function, absolute_import, division
import sys
import subprocess
from distutils.spawn import find_executable
from setuptools import setup, find_packages
from basesetup import write_version_py

VERSION = '1.0.0'
ISRELEASED = True
__version__ = VERSION

def main(**kwargs):
    write_version_py(VERSION, ISRELEASED, 'osprey/version.py')

    classifiers = """\
    Development Status :: 3 - Alpha
    Intended Audience :: Science/Research
    License :: OSI Approved :: Apache Software License
    Programming Language :: Python
    Programming Language :: Python :: 2.7
    Programming Language :: Python :: 3
    Programming Language :: Python :: 3.4
    Programming Language :: Python :: 3.5
    Operating System :: Unix
    Operating System :: MacOS
    Operating System :: Microsoft :: Windows
    Topic :: Scientific/Engineering
    Topic :: Scientific/Engineering :: Information Analysis"""
    setup(
        name='osprey',
        author='Robert T. McGibbon',
        author_email='rmcgibbo@gmail.com',
        url='https://github.com/msmbuilder/osprey',
        classifiers=[e.strip() for e in classifiers.splitlines()],
        platforms=["Windows", "Linux", "Mac OS-X", "Unix"],
        license='Apache Software License',
        download_url='https://pypi.python.org/pypi/osprey/',
        version=VERSION,
        packages=find_packages(),
        zip_safe=False,
        package_data={'osprey': ['data/*']},
        entry_points={
            'console_scripts': [
                'osprey = osprey.cli.main:main',
            ],
        },
        **kwargs
    )


def readme_to_rst():
    pandoc = find_executable('pandoc')
    if pandoc is None:
        raise RuntimeError("Turning the readme into a description requires "
                           "pandoc.")
    long_description = subprocess.check_output(
        [pandoc, 'README.md', '-t', 'rst'])
    short_description = long_description.split('\n\n')[1]
    return {
        'description': short_description,
        'long_description': long_description,
    }


if __name__ == '__main__':
    kwargs = {}
    if any(e in sys.argv for e in ('upload', 'register', 'sdist')):
        kwargs = readme_to_rst()
    main(**kwargs)



from __future__ import print_function, absolute_import, division
import os
from os.path import samefile, abspath
import tempfile
import datetime
from osprey.utils import dict_merge, in_directory
from osprey.utils import format_timedelta, current_pretty_time


def test_dict_merge_1():
    base = {'a': 1}
    top = {'a': 2}
    assert dict_merge(base, top) == top


def test_dict_merge_2():
    base = {'a': 1}
    top = {'a': 2, 'b': 3}
    assert dict_merge(base, top) == top


def test_dict_merge_3():
    base = {'a': 1, 'b': 2}
    top = {'a': 2}
    assert dict_merge(base, top) == {'a': 2, 'b': 2}


def test_dict_merge_4():
    base = {'a': {'a': 1}, 'b': {'b': 2}}
    top = {'a': {'a': 2, 'b': 3}}
    assert dict_merge(base, top) == {'a': {'a': 2, 'b': 3}, 'b': {'b': 2}}


def test_in_directory_1():
    tempdir = tempfile.mkdtemp()
    try:
        initialdir = abspath(os.curdir)
        with in_directory(tempdir):
            assert samefile(abspath(os.curdir), tempdir)
        assert samefile(abspath(os.curdir), initialdir)
    finally:
        os.rmdir(tempdir)


def test_format_timedelta():
    print(format_timedelta(datetime.timedelta(seconds=413302.33)))


def test_current_pretty_time():
    print(current_pretty_time())



from __future__ import print_function, absolute_import, division
import numpy as np
import os
import tempfile

from six.moves import cPickle
from sklearn.cluster import KMeans

from osprey.config import Config
from osprey.search_space import IntVariable, FloatVariable, EnumVariable
from osprey.strategies import RandomSearch, HyperoptTPE, GP


os.environ['OSPREYRC'] = ' '


def test_estimator_pickle():
    with tempfile.NamedTemporaryFile('w+b', 0) as f:

        cPickle.dump(KMeans(), f)

        config = Config.fromdict({
            'estimator': {'pickle': f.name}
        }, check_fields=False)
        assert isinstance(config.estimator(), KMeans)


def test_estimator_eval_1():
    config = Config.fromdict({
        'estimator': {
            'eval': 'KMeans()',
            'eval_scope': 'sklearn',
        }
    }, check_fields=False)
    assert isinstance(config.estimator(), KMeans)


def test_estimator_eval_2():
    config = Config.fromdict({
        'estimator': {
            'eval': 'KMeans()',
            'eval_scope': ['sklearn'],
        }
    }, check_fields=False)
    assert isinstance(config.estimator(), KMeans)


def test_estimator_entry_point():
    config = Config.fromdict({
        'estimator': {
            'entry_point': 'sklearn.cluster.KMeans',
        }
    }, check_fields=False)
    assert isinstance(config.estimator(), KMeans)


def test_estimator_entry_point_params():
    config = Config.fromdict({
        'estimator': {
            'entry_point': 'sklearn.cluster.KMeans',
            'params': {
                'n_clusters': 15
            }
        }
    }, check_fields=False)
    assert isinstance(config.estimator(), KMeans)
    assert config.estimator().n_clusters == 15


def test_search_space():
    config = Config.fromdict({
        'search_space': {
            'intvar': {'type': 'int', 'min': 1, 'max': 2},
            'logivar': {'type': 'int', 'min': 1, 'max': 2, 'warp': 'log'},
            'fvar': {'type': 'float', 'min': 1, 'max': 3.5},
            'logfvar': {'type': 'float', 'min': 1, 'max': 2.5, 'warp': 'log'},
            'enumvar': {'type': 'enum', 'choices': [1, False]},
            'jumpvar': {'type': 'jump',  'min': 1, 'max': 3, 'step': 1,'var_type':int}
        }}, check_fields=False)
    searchspace = config.search_space()
    assert searchspace['intvar'] == IntVariable('intvar', 1, 2, warp=None)
    assert searchspace['logivar'] == IntVariable('logivar', 1, 2, warp='log')
    assert searchspace['fvar'] == FloatVariable('fvar', 1, 3.5, warp=None)
    assert searchspace['logfvar'] == FloatVariable('logfvar', 1, 2.5,
                                                   warp='log')
    assert searchspace['enumvar'] == EnumVariable('enumvar', [1, False])
    assert searchspace['jumpvar'] == EnumVariable('jumpvar', [1.0, 2.0, 3.0])


def test_strategy_random():
    config = Config.fromdict({
        'strategy': {'name': 'random'}
    }, check_fields=False)
    assert isinstance(config.strategy(), RandomSearch)


def test_search_engine_hyperopt_tpe():
    config = Config.fromdict({
        'strategy': {'name': 'hyperopt_tpe'}
    }, check_fields=False)
    assert isinstance(config.strategy(), HyperoptTPE)


def test_search_engine_gp():
    config = Config.fromdict({
        'strategy': {'name': 'gp'}
    }, check_fields=False)
    assert isinstance(config.strategy(), GP)


def test_scoring():
    config = Config.fromdict({
        'scoring': 'sdfsfsdf'
    }, check_fields=False)
    assert config.scoring() is 'sdfsfsdf'


def test_cv_1():
    from sklearn.cross_validation import ShuffleSplit
    for name in ['shufflesplit', 'ShuffleSplit']:
        config = Config.fromdict({
            'cv': {'name': name, 'params': {'n_iter': 10}}
        }, check_fields=False)
        cv = config.cv(range(100))
        assert isinstance(cv, ShuffleSplit)
        assert cv.n_iter == 10


def test_stratified_cv():
    from sklearn.cross_validation import StratifiedShuffleSplit
    config = Config.fromdict({
        'cv': {'name': 'stratifiedshufflesplit', 'params': {'n_iter': 10}}
    }, check_fields=False)
    cv = config.cv(range(100), np.random.randint(2, size=100))
    assert isinstance(cv, StratifiedShuffleSplit)
    assert cv.n_iter == 10



from __future__ import print_function, absolute_import, division

from .utils import num_samples

import numpy as np


class BaseCVFactory(object):
    short_name = None

    def load(self):
        raise NotImplementedError('should be implemented in subclass')

    def create(self, X, y):
        raise NotImplementedError('should be implemented in subclass')


class ShuffleSplitFactory(BaseCVFactory):
    short_name = ['shufflesplit', 'ShuffleSplit']

    def __init__(self, n_iter=10, test_size=0.1, train_size=None,
                 random_state=None):
        self.n_iter = n_iter
        self.test_size = test_size
        self.train_size = train_size
        self.random_state = random_state

    def create(self, X, y=None):
        from sklearn.cross_validation import ShuffleSplit

        return ShuffleSplit(num_samples(X), n_iter=self.n_iter,
                            test_size=self.test_size,
                            train_size=self.train_size,
                            random_state=self.random_state)


class KFoldFactory(BaseCVFactory):
    short_name = ['kfold', 'KFold']

    def __init__(self, n_folds=3, shuffle=False, random_state=None):
        self.n_folds = n_folds
        self.shuffle = shuffle
        self.random_state = random_state

    def create(self, X, y=None):
        from sklearn.cross_validation import KFold

        return KFold(num_samples(X), n_folds=self.n_folds, shuffle=self.shuffle,
                     random_state=self.random_state)


class LeaveOneOutFactory(BaseCVFactory):
    short_name = ['loo', 'LeaveOneOut']

    def __init__(self):
        pass

    def create(self, X, y=None):
        from sklearn.cross_validation import LeaveOneOut

        return LeaveOneOut(num_samples(X))


class StratifiedShuffleSplitFactory(BaseCVFactory):
    short_name = ['stratifiedshufflesplit', 'StratifiedShuffleSplit']

    def __init__(self, n_iter=10, test_size=0.1, train_size=None,
                 random_state=None):
        self.n_iter = n_iter
        self.test_size = test_size
        self.train_size = train_size
        self.random_state = random_state

    def create(self, X, y):
        from sklearn.cross_validation import StratifiedShuffleSplit

        return StratifiedShuffleSplit(y, n_iter=self.n_iter,
                                      test_size=self.test_size,
                                      train_size=self.train_size,
                                      random_state=self.random_state)


class StratifiedKFoldFactory(BaseCVFactory):
    short_name = ['stratifiedkfold', 'StratifiedKFold']

    def __init__(self, n_folds=3, shuffle=False, random_state=None):
        self.n_folds = n_folds
        self.shuffle = shuffle
        self.random_state = random_state

    def create(self, X, y):
        from sklearn.cross_validation import StratifiedKFold

        return StratifiedKFold(y, n_folds=self.n_folds, shuffle=self.shuffle,
                               random_state=self.random_state)


class FixedCVFactory(BaseCVFactory):
    """
    Cross-validator to use with a fixed, held-out validation set.

    Parameters
    ----------
    start : int
        Start index of validation set.
    stop : int, optional
        Stop index of validation set.
    """
    short_name = ['fixed', 'Fixed']

    def __init__(self, start, stop=None):
        self.valid = slice(start, stop)

    def create(self, X, y):
        indices = np.arange(num_samples(X))
        valid = indices[self.valid]
        train = np.setdiff1d(indices, valid)
        return (train, valid),  # return a nested tuple



from __future__ import print_function, absolute_import, division
from argparse import ArgumentDefaultsHelpFormatter


def func(args, parser):
    # delay import of the rest of the module to improve `osprey -h` performance
    from ..execute_skeleton import execute
    execute(args, parser)


def configure_parser(sub_parsers):
    help = 'create skeleton config.yaml file'
    p = sub_parsers.add_parser('skeleton', description=help, help=help,
                               formatter_class=ArgumentDefaultsHelpFormatter)
    p.add_argument('-t', '--template', help=(
        "which skeleton to create. 'msmbuilder' is a skeleton config file for"
        "MSMBuilder molecular dynamics / Markov state model based "
        "projects."), choices=['msmbuilder', 'sklearn'], default='msmbuilder',)
    p.add_argument('-f', '--filename', help='config filename to create',
                   default='config.yaml')
    p.set_defaults(func=func)



from __future__ import print_function, absolute_import, division

import sys
import argparse

from .. import __version__
from . import parser_dump
from . import parser_skeleton
from . import parser_worker
from . import parser_plot
from . import parser_currentbest


def main():
    help = 'osprey is a tool for machine learning hyperparameter optimization.'
    p = argparse.ArgumentParser(description=help)
    p.add_argument(
        '-V', '--version',
        action='version',
        version='osprey %s' % __version__,
    )
    sub_parsers = p.add_subparsers(
        metavar='command',
        dest='cmd',
    )

    parser_dump.configure_parser(sub_parsers)
    parser_skeleton.configure_parser(sub_parsers)
    parser_worker.configure_parser(sub_parsers)
    parser_plot.configure_parser(sub_parsers)
    parser_currentbest.configure_parser(sub_parsers)

    if len(sys.argv) == 1:
        sys.argv.append('-h')

    args = p.parse_args()
    args_func(args, p)


def args_func(args, p):
    try:
        args.func(args, p)
    except RuntimeError as e:
        sys.exit("Error: %s" % e)
    except Exception as e:
        if e.__class__.__name__ not in ('ScannerError', 'ParserError'):
            message = """\
An unexpected error has occurred with osprey (version %s), please
consider sending the following traceback to the osprey GitHub issue tracker at:
        https://github.com/pandegroup/osprey/issues
"""
            print(message % __version__, file=sys.stderr)
        raise  # as if we did not catch it



from __future__ import print_function, absolute_import, division
import numpy as np
import re
from string import Template

from sklearn.base import BaseEstimator
from sklearn.metrics import accuracy_score, mean_squared_error

from osprey.dataset_loaders import BaseDatasetLoader


class Pylearn2Estimator(BaseEstimator):
    """
    Wrapper for pylearn2 models that conforms to the sklearn BaseEstimator API.

    This class does not handle Train objects directly, because it is much
    simpler to handle set_params with the YAML and simple string formatting.

    YAML strings should be formatted for compatibility with string.Template.

    Parameters
    ----------
    yaml_string : str
        YAML string defining the model.
    """
    def __init__(self, yaml_string, **kwargs):
        self.trainer = None
        self.yaml_string = yaml_string
        self.set_params(**kwargs)

    def _get_param_names(self):
        """
        Get mappable parameters from YAML.
        """
        template = Template(self.yaml_string)
        names = ['yaml_string']  # always include the template
        for match in re.finditer(template.pattern, template.template):
            name = match.group('named') or match.group('braced')
            assert name is not None
            names.append(name)
        return names

    def _get_dataset(self, X, y=None):
        """
        Construct a pylearn2 dataset.

        Parameters
        ----------
        X : array_like
            Training examples.
        y : array_like, optional
            Labels.
        """
        from pylearn2.datasets import DenseDesignMatrix

        X = np.asarray(X)
        assert X.ndim > 1
        if y is not None:
            y = self._get_labels(y)
        if X.ndim == 2:
            return DenseDesignMatrix(X=X, y=y)
        return DenseDesignMatrix(topo_view=X, y=y)

    def _get_labels(self, y):
        """
        Construct pylearn2 dataset labels.

        Parameters
        ----------
        y : array_like, optional
            Labels.
        """
        y = np.asarray(y)
        if y.ndim == 1:
            return y.reshape((y.size, 1))
        assert y.ndim == 2
        return y

    def fit(self, X, y=None):
        """
        Build a trainer and run main_loop.

        Parameters
        ----------
        X : array_like
            Training examples.
        y : array_like, optional
            Labels.
        """
        from pylearn2.config import yaml_parse
        from pylearn2.train import Train

        # build trainer
        params = self.get_params()
        yaml_string = Template(self.yaml_string).substitute(params)
        self.trainer = yaml_parse.load(yaml_string)
        assert isinstance(self.trainer, Train)
        if self.trainer.dataset is not None:
            raise ValueError('Train YAML database must evaluate to None.')
        self.trainer.dataset = self._get_dataset(X, y)

        # update monitoring dataset(s)
        if (hasattr(self.trainer.algorithm, 'monitoring_dataset') and
                self.trainer.algorithm.monitoring_dataset is not None):
            monitoring_dataset = self.trainer.algorithm.monitoring_dataset
            if len(monitoring_dataset) == 1 and '' in monitoring_dataset:
                monitoring_dataset[''] = self.trainer.dataset
            else:
                monitoring_dataset['train'] = self.trainer.dataset
            self.trainer.algorithm._set_monitoring_dataset(monitoring_dataset)
        else:
            self.trainer.algorithm._set_monitoring_dataset(
                self.trainer.dataset)

        # run main loop
        self.trainer.main_loop()

    def predict(self, X):
        """
        Get model predictions.

        Parameters
        ----------
        X : array_like
            Test dataset.
        """
        return self._predict(X)

    def _predict(self, X, method='fprop'):
        """
        Get model predictions.

        See pylearn2.scripts.mlp.predict_csv and
        http://fastml.com/how-to-get-predictions-from-pylearn2/.

        Parameters
        ----------
        X : array_like
            Test dataset.
        method : str
            Model method to call for prediction.
        """
        import theano

        X_sym = self.trainer.model.get_input_space().make_theano_batch()
        y_sym = getattr(self.trainer.model, method)(X_sym)
        f = theano.function([X_sym], y_sym, allow_input_downcast=True)
        return f(X)

    def score(self, X, y):
        """
        Score predictions.

        Parameters
        ----------
        X : array_like
            Test examples.
        y : array_like, optional
            Labels.
        """
        raise NotImplementedError('should be implemented in subclasses')


class Pylearn2Classifier(Pylearn2Estimator):
    """
    pylearn2 classifier.
    """
    def _get_labels(self, y):
        """
        Construct pylearn2 dataset labels.

        Parameters
        ----------
        y : array_like, optional
            Labels.
        """
        y = np.asarray(y)
        assert y.ndim == 1
        # convert to one-hot
        labels = np.unique(y).tolist()
        oh = np.zeros((y.size, len(labels)), dtype=float)
        for i, label in enumerate(y):
            oh[i, labels.index(label)] = 1.
        return oh

    def predict_proba(self, X):
        """
        Get model predictions.

        Parameters
        ----------
        X : array_like
            Test dataset.
        """
        return self._predict(X)

    def predict(self, X):
        """
        Get model predictions.

        Parameters
        ----------
        X : array_like
            Test dataset.
        """
        return np.argmax(self._predict(X), axis=1)

    def score(self, X, y):
        """
        Score predictions.

        Parameters
        ----------
        X : array_like
            Test examples.
        y : array_like, optional
            Labels.
        """
        return accuracy_score(y, self.predict(X))


class Pylearn2Regressor(Pylearn2Estimator):
    """
    pylearn2 regressor.
    """
    def score(self, X, y):
        """
        Score predictions.

        Parameters
        ----------
        X : array_like
            Test examples.
        y : array_like, optional
            Labels.
        """
        return mean_squared_error(y, self.predict(X))


class Pylearn2Autoencoder(Pylearn2Estimator):
    """
    pylearn2 autoencoder.
    """
    def _predict(self, X, method='reconstruct'):
        return super(Pylearn2Autoencoder, self)._predict(X, method)

    def score(self, X, y):
        """
        Score predictions.

        Parameters
        ----------
        X : array_like
            Test examples.
        y : array_like, optional
            Labels (not used).
        """
        return mean_squared_error(X, self.predict(X))


class Pylearn2DatasetLoader(BaseDatasetLoader):
    """
    pylearn2 dataset loader.

    Parameters
    ----------
    yaml_string : str
        YAML specification of a pylearn2 dataset.
    one_hot : bool, optional
        Take argmax of one-hot labels to get classes.
    """
    short_name = 'pylearn2_dataset'

    def __init__(self, yaml_string, one_hot=False):
        self.yaml_string = yaml_string
        self.one_hot = one_hot

    def load(self):
        """
        Load the dataset using pylearn2.config.yaml_parse.
        """
        from pylearn2.config import yaml_parse
        from pylearn2.datasets import Dataset

        dataset = yaml_parse.load(self.yaml_string)
        assert isinstance(dataset, Dataset)
        data = dataset.iterator(mode='sequential', num_batches=1,
                                data_specs=dataset.data_specs,
                                return_tuple=True).next()
        if len(data) == 2:
            X, y = data
            y = np.squeeze(y)
            if self.one_hot:
                y = np.argmax(y, axis=1)
        else:
            X = data
            y = None
        return X, y



from __future__ import print_function, absolute_import, division

import time
from distutils.version import LooseVersion

import numpy as np
import sklearn
from sklearn.base import is_classifier, clone
from sklearn.metrics.scorer import check_scoring
from sklearn.externals.joblib import Parallel, delayed
from sklearn.cross_validation import check_cv, _safe_split, _score

from .utils import check_arrays, num_samples
from .utils import short_format_time, is_msmbuilder_estimator


if LooseVersion(sklearn.__version__) < LooseVersion('0.16.1'):
    raise ImportError('Please upgrade to the latest version of scikit-learn')


def fit_and_score_estimator(estimator, parameters, cv, X, y=None, scoring=None,
                            iid=True, n_jobs=1, verbose=1,
                            pre_dispatch='2*n_jobs'):
    """Fit and score an estimator with cross-validation

    This function is basically a copy of sklearn's
    grid_search._BaseSearchCV._fit(), which is the core of the GridSearchCV
    fit() method. Unfortunately, that class does _not_ return the training
    set scores, which we want to save in the database, and because of the
    way it's written, you can't change it by subclassing or monkeypatching.

    This function uses some undocumented internal sklearn APIs (non-public).
    It was written against sklearn version 0.16.1. Prior Versions are likely
    to fail due to changes in the design of cross_validation module.

    Returns
    -------
    out : dict, with keys 'mean_test_score' 'test_scores', 'train_scores'
        The scores on the training and test sets, as well as the mean test set
        score.
    """
    scorer = check_scoring(estimator, scoring=scoring)
    n_samples = num_samples(X)
    X, y = check_arrays(X, y, allow_lists=True, sparse_format='csr',
                        allow_nans=True)
    if y is not None:
        if len(y) != n_samples:
            raise ValueError('Target variable (y) has a different number '
                             'of samples (%i) than data (X: %i samples)'
                             % (len(y), n_samples))
    cv = check_cv(cv, X, y, classifier=is_classifier(estimator))

    out = Parallel(
        n_jobs=n_jobs, verbose=verbose, pre_dispatch=pre_dispatch
    )(
        delayed(_fit_and_score)(clone(estimator), X, y, scorer,
                                train, test, verbose, parameters,
                                fit_params=None)
        for train, test in cv)

    assert len(out) == len(cv)

    train_scores, test_scores = [], []
    n_train_samples, n_test_samples = [], []
    for test_score, n_test, train_score, n_train, _ in out:
        train_scores.append(train_score)
        test_scores.append(test_score)
        n_test_samples.append(n_test)
        n_train_samples.append(n_train)

    train_scores, test_scores = map(list, check_arrays(train_scores,
                                                       test_scores,
                                                       warn_nans=True,
                                                       replace_nans=True))

    if iid:
        if verbose > 0 and is_msmbuilder_estimator(estimator):
            print('[CV] Using MSMBuilder API n_samples averaging')
            print('[CV]   n_train_samples: %s' % str(n_train_samples))
            print('[CV]   n_test_samples: %s' % str(n_test_samples))
        mean_test_score = np.average(test_scores, weights=n_test_samples)
        mean_train_score = np.average(train_scores, weights=n_train_samples)
    else:
        mean_test_score = np.average(test_scores)
        mean_train_score = np.average(train_scores)

    grid_scores = {
        'mean_test_score': mean_test_score, 'test_scores': test_scores,
        'mean_train_score': mean_train_score, 'train_scores': train_scores,
        'n_test_samples': n_test_samples, 'n_train_samples': n_train_samples}
    return grid_scores


def _fit_and_score(estimator, X, y, scorer, train, test, verbose, parameters,
                   fit_params=None):
    if verbose > 1:
        if parameters is None:
            msg = "no parameters to be set"
        else:
            msg = '%s' % (', '.join('%s=%s' % (k, v)
                          for k, v in parameters.items()))
        print("[CV] %s %s" % (msg, (64 - len(msg)) * '.'))

    if num_samples(train) == 0 or num_samples(test) == 0:
        raise RuntimeError(
            'Cross validation error in fit_estimator. The total data set '
            'contains %d elements, which were split into a training set '
            'of %d elements and a test set of %d elements. Unfortunately, '
            'you can\'t have a %s set with 0 elements.' % (
                num_samples(X), num_samples(train), num_samples(test),
                'training' if num_samples(train) == 0 else 'test'))

    # adjust length of sample weights
    n_samples = num_samples(X)
    fit_params = fit_params if fit_params is not None else {}
    fit_params = dict([(k, np.asarray(v)[train]
                       if hasattr(v, '__len__') and len(v) == n_samples else v)
                       for k, v in fit_params.items()])

    if parameters is not None:
        estimator.set_params(**parameters)

    # fit and score
    start_time = time.time()

    X_train, y_train = _safe_split(estimator, X, y, train)
    X_test, y_test = _safe_split(estimator, X, y, test, train)
    if y_train is None:
        estimator.fit(X_train, **fit_params)
    else:
        estimator.fit(X_train, y_train, **fit_params)
    test_score = _score(estimator, X_test, y_test, scorer)
    train_score = _score(estimator, X_train, y_train, scorer)

    scoring_time = time.time() - start_time

    msmbuilder_api = is_msmbuilder_estimator(estimator)
    n_samples_test = num_samples(X_test, is_nested=msmbuilder_api)
    n_samples_train = num_samples(X_train, is_nested=msmbuilder_api)
    if verbose > 2:
        msg += ", score=%f" % test_score
    if verbose > 1:
        end_msg = "%s -%s" % (msg, short_format_time(scoring_time))
        print("[CV] %s %s" % ((64 - len(end_msg)) * '.', end_msg))

    return (test_score, n_samples_test, train_score, n_samples_train,
            scoring_time)



from __future__ import print_function, absolute_import, division

import sys
import traceback
import importlib


def load_entry_point(entry_point_str, section=''):
    try:
        package_str, obj_str = entry_point_str.rsplit('.', 1)
    except ValueError:
        raise RuntimeError('%s syntax error' % section)

    try:
        package = importlib.import_module(package_str)
    except ImportError:
        print('-'*78)
        print('Error parsing %s' % section, file=sys.stderr)
        print('-'*78)
        traceback.print_exc(sys.stderr)
        sys.exit(1)

    try:
        value = getattr(package, obj_str)
    except (AttributeError, KeyError):
        raise RuntimeError('%s: %r does not contain object %r' % (
            section, package_str, obj_str))

    return value
